#include "3d.h"  
#include "hair.h"
//#include "trig.h"

#define INT(x) ((int)(x))
#define SQR(x) ((x)*(x))

#define InBounds( x, bounds )   ( ( (x) >= 0 ) && ( (x) < (bounds) ) )
#define MakeInBounds( x, bounds )   if ( x < 0  ) x = 0; if( x >= (bounds) ) x = (bounds)-1

/********* things that may cause problems ******************
* the function FindZ() was made into a macro that doesn't  *
* do error checking.                                       *
*----------------------------------------------------------*
***********************************************************/

/********************** NOTES *******************************
*  For a left-handed system?? (counter-clockwise polygons): *
*         make sure InitZBuffer() initilizes the zbuffer to *
*       a very small (negative) number.                     *
*  Also when making compares against current z (in scanfill)*
* make sure it is if( z > temp_z )                          *
*-----------------------------------------------------------*
* Both should be opposite for (clockwise system) (obj files *
* outputted from commericial programs)                      *
************************************************************/

long cur_vertex_id = 1;
long cur_normal_id = 1;

char *shademap;

int *z_buffer;
int zbuf_rows, zbuf_cols;

Vertex *light=NULL;

/* doesn't work */
void Object_3d :: Tween( Object_3d *object, float t ){
	MoveFirst();
	object->MoveFirst();

	while( face != NULL ){
		face->face->MoveFirst();
		object->face->face->MoveFirst();

		while( face->face->vertex != NULL ){
			face->face->vertex->x = (t*(object->face->face->vertex->x - face->face->vertex->x)) + face->face->vertex->x;
			face->face->vertex->y = (t*(object->face->face->vertex->y - face->face->vertex->y)) + face->face->vertex->y;
			face->face->vertex->z = (t*(object->face->face->vertex->z - face->face->vertex->z)) + face->face->vertex->z;

			face->face->MoveNext();
			object->face->face->MoveNext();
		}

		face->face->MoveFirst();
		object->face->face->MoveFirst();

		MoveNext();
		object->MoveNext();
	}

	MoveFirst();
	object->MoveFirst();
}

void Init3DEngine( int xsize, int ysize ){
	int dummy;
	paletteType pal;

	light = NewVertex( 0.0, 0.0, -1.0 );
	Normalize( light );

	zbuf_cols = xsize;
	zbuf_rows = ysize;
	z_buffer = new int[ xsize*ysize ];

	InitZBuffer();

	/*** Load the shade map for Fast Phong Shading ***/
	shademap = new char[65536];   /* 256 x 256 */
	LoadPCX( "shademap.pcx", (unsigned char **)&shademap, &pal, &dummy, &dummy );
}

void Object_3d :: CalcPhongMap( void ){
	int i, color;

	if( phongmap == NULL )
		phongmap = new char[65536];

	for(i=0; i<65536; ++i ){
		color = (int)( ( (float)((unsigned char)shademap[i])/255.0)*(high_color-low_color) ) + low_color;
		if( color < low_color ) color = low_color;
		if( color > high_color ) color = high_color;
		phongmap[i] = (char)color;
	} 
}

void Face :: ChangeColor( int low, int high ){
	MakeInBounds( low_color, 255 );
	MakeInBounds( high_color, 255 );

	low_color = low;
	high_color = high;
}

void Object_3d :: ChangeColor( int low, int high ){
	MoveFirst();

	MakeInBounds( low_color, 255 );
	MakeInBounds( high_color, 255 );

	low_color = low;
	high_color = high;

	while( face != NULL ){
		face->face->low_color = low;
		face->face->high_color = high;

		MoveNext();
	}

	MoveFirst();

	if( use_fastphong_shade )
		CalcPhongMap();
}

Camera *MakeCamera( void ){
        Camera *temp;
 
        temp = new Camera;
 
        temp->position = NewVertex( .5, .5,  1 );
        temp->focus    = NewVertex( .5, .5, .5 );

	temp->u = NewVertex( 1, 0, 0 );
	temp->v = NewVertex( 0, 1, 0 );
	temp->n = NewVertex( 0, 0, 1 );

	return temp;
}

void Object_3d :: AddCamera( Camera *camera ){
}
void  AddCameraToHair( Camera *camera, Hair *hair ){
	matrix_type xyzw;
	matrix_type *result;
	matrix_type m;
	static Vertex *u=NULL, *v, *n;
	float magnitude;
	int i, j, k;

	xyzw.rows = 4;
	xyzw.cols = 1;

	magnitude = sqrt( SQR( camera->focus->x - camera->position->x ) + SQR( camera->focus->y - camera->position->y ) + SQR( camera->focus->z - camera->position->z ) ); 

	if( u == NULL ){
		u = NewVertex( 1, 0, 0 );
		v = NewVertex( 0, 1, 0 );
		n = NewVertex( 0, 0, 1 );
	}

	n->x = -camera->focus->x + camera->position->x;
	n->y = -camera->focus->y + camera->position->y;
	n->z = -camera->focus->z + camera->position->z;

	if( n->y != 0.0 ){
		if( n->y > 0 ){
		v->x = -n->x;
		v->z = -n->z;
		}
		else{
		v->x = n->x;
		v->z = n->z;
		}
		v->y = ( (-1.0 * n->x * v->x) - (v->z * n->z) ) / n->y;
	}
	else if ( n->z != 0.0 ){
		v->x = 0;
		v->y = 1.0;
		v->z = ( (-1.0 * n->x * v->x) - (v->y * n->y) ) / n->z;
	}
	else if ( n->x != 0.0 ){
		v->y = 1.0;
		v->z = 0;
		v->x = ( (-1.0 * n->y * v->y) - (v->z * n->z) ) / n->x;
	}else{
		fprintf( stderr, "Camera position = camera focus: cannot handle this setup!\n" );
		exit( 0 );
	}
	
	u->x = (v->y * n->z) - (v->z * n->y);
	u->y = (v->z * n->x) - (v->x * n->z);
	u->z = (v->x * n->y) - (v->y * n->x);

	Normalize( u );
	Normalize( v );
	Normalize( n );

	m.rows = m.cols = 4;

	m.matrix[0][0] = u->x;
	m.matrix[0][1] = u->y;
	m.matrix[0][2] = u->z;
	m.matrix[0][3] = 0;

	m.matrix[1][0] = v->x;
	m.matrix[1][1] = v->y;
	m.matrix[1][2] = v->z;
	m.matrix[1][3] = 0;

	m.matrix[2][0] = n->x;
	m.matrix[2][1] = n->y;
	m.matrix[2][2] = n->z;
	m.matrix[2][3] = 0;

	m.matrix[3][0] = 0;
	m.matrix[3][1] = 0;
	m.matrix[3][2] = 0;
	m.matrix[3][3] = 1;

	for(i=0; i<NUM_HAIRS; ++i )
		for(j=0; j<LENGTH_SEGMENTS; ++j ){
			hair->strands[ i ].joints[ j ].x -= camera->position->x;
			hair->strands[ i ].joints[ j ].y -= camera->position->y;
			hair->strands[ i ].joints[ j ].z -= camera->position->z;

			for(k=0; k<WIDTH_SEGMENTS; ++k ){
				hair->strands[ i ].joints[ j ].points[ k ].x -= camera->position->x;
				hair->strands[ i ].joints[ j ].points[ k ].y -= camera->position->y;
				hair->strands[ i ].joints[ j ].points[ k ].z -= camera->position->z;
			}
		}

	for(i=0; i<NUM_HAIRS; ++i )
		for(j=0; j<LENGTH_SEGMENTS; ++j ){
			xyzw.matrix[0][0] = hair->strands[ i ].joints[ j ].x;
        		xyzw.matrix[1][0] = hair->strands[ i ].joints[ j ].y;
                	xyzw.matrix[2][0] = hair->strands[ i ].joints[ j ].z;
           		xyzw.matrix[3][0] = 0.0;

	               	result = MatrixMultiply4x4_X_4x1( &m, &xyzw );
	
       	        	hair->strands[ i ].joints[ j ].x = (float)result->matrix[0][0];
               		hair->strands[ i ].joints[ j ].y = (float)result->matrix[1][0];
               		hair->strands[ i ].joints[ j ].z = (float)result->matrix[2][0];

			for(k=0; k<WIDTH_SEGMENTS; ++k ){
		                xyzw.matrix[0][0] = hair->strands[ i ].joints[ j ].points[ k ].x;
        		        xyzw.matrix[1][0] = hair->strands[ i ].joints[ j ].points[ k ].y;
                		xyzw.matrix[2][0] = hair->strands[ i ].joints[ j ].points[ k ].z;
           			xyzw.matrix[3][0] = 0.0;

	                	result = MatrixMultiply4x4_X_4x1( &m, &xyzw );
	
        	        	hair->strands[ i ].joints[ j ].points[ k ].x = (float)result->matrix[0][0];
                		hair->strands[ i ].joints[ j ].points[ k ].y = (float)result->matrix[1][0];
                		hair->strands[ i ].joints[ j ].points[ k ].z = (float)result->matrix[2][0];
			}
		}
	

	for(i=0; i<NUM_HAIRS; ++i )
		for(j=0; j<LENGTH_SEGMENTS; ++j ){
			hair->strands[ i ].joints[ j ].x += 0.5;
			hair->strands[ i ].joints[ j ].y += 0.5;
			hair->strands[ i ].joints[ j ].z += 0.5-magnitude;

			for(k=0; k<WIDTH_SEGMENTS; ++k ){
				hair->strands[ i ].joints[ j ].points[ k ].x += 0.5;
				hair->strands[ i ].joints[ j ].points[ k ].y += 0.5;
				hair->strands[ i ].joints[ j ].points[ k ].z += 0.5-magnitude;
			}
		}
}


/*******************************************************************
*  This function is VERY important to call before any solid images *
* are drawn to the screen.                                         *
*******************************************************************/
void InitZBuffer( void ){
	int temp = zbuf_cols*zbuf_rows;

        for(int i=0; i<temp; ++i )
                z_buffer[i] = -102400000;
}

void Face :: AddCamera( matrix_type *m ){
	matrix_type xyzw;
	matrix_type *result;

	xyzw.rows = 4;
	xyzw.cols = 1;

        MoveFirst();

        while( vertex != NULL ){
                xyzw.matrix[0][0] = vertex->x;
                xyzw.matrix[1][0] = vertex->y;
                xyzw.matrix[2][0] = vertex->z;
                xyzw.matrix[3][0] = vertex->w;

                result = MatrixMultiply4x4_X_4x1( m, &xyzw );

                vertex->x = (float)result->matrix[0][0];
                vertex->y = (float)result->matrix[1][0];
                vertex->z = (float)result->matrix[2][0];
                vertex->w = (float)result->matrix[3][0];

                xyzw.matrix[0][0] = vertex->normal->x;
                xyzw.matrix[1][0] = vertex->normal->y;
                xyzw.matrix[2][0] = vertex->normal->z;
                xyzw.matrix[3][0] = 0;

                result = MatrixMultiply4x4_X_4x1( m, &xyzw );

                vertex->normal->x = (float)result->matrix[0][0];
                vertex->normal->y = (float)result->matrix[1][0];
                vertex->normal->z = (float)result->matrix[2][0];
                vertex->normal->w = (float)result->matrix[3][0];

                MoveNext();
	}

        xyzw.matrix[0][0] = normal_to_face->x;
        xyzw.matrix[1][0] = normal_to_face->y;
        xyzw.matrix[2][0] = normal_to_face->z;
        xyzw.matrix[3][0] = normal_to_face->w;

        result = MatrixMultiply4x4_X_4x1( m, &xyzw );

        normal_to_face->x = (float)result->matrix[0][0];
        normal_to_face->y = (float)result->matrix[1][0];
        normal_to_face->z = (float)result->matrix[2][0];
        normal_to_face->w = (float)result->matrix[3][0];

	MoveFirst();
}



Face :: Face( void ){
        low_color = 0;
        high_color = 255;
        begin = vertex = end = NULL;
        begin_per = in_perspective = end_per = NULL;

        use_perspective = 0.0;
        remove_backfaces = 0;
}

Face :: ~Face( void ){
        /* free memory */
}

void Face :: Scale( float amount ){
	MoveFirst();

	while( vertex != NULL ){
		vertex->x = (vertex->x * amount);
		vertex->y = (vertex->y * amount);
		vertex->z = (vertex->z * amount);

		MoveNext();
	}

	MoveFirst();
}

void Object_3d :: Scale( float amount ){
	MoveFirst();

	while( face != NULL ){
		face->face->Scale( amount );

		MoveNext();
	}

	MoveFirst();
}

char Face :: MoveToVertex( int id ){
        char result = 0;

        MoveFirst();

        while( vertex != NULL ){
                if( vertex->id == id ){
                        result = 1;
                        break;
                }else{
                        vertex = vertex->next;
                }
        }

        return( result );
}

void Face :: UsePerspective( float v ){
        use_perspective = v;
}

void Face :: RemoveBackfaces( char v ){
        remove_backfaces = v;
}


char Face :: MoveToNormal( int id ){
        char result = 0;

        MoveFirst();

        while( vertex != NULL ){
                if( vertex->normal != NULL ){
                        if( vertex->normal->id == id ){
                                result = 1;
                                break;
                        }else{
                                vertex = vertex->next;
                        }
                }
        }

        return( result );       
}

void Face :: MoveFirst( void ){
        vertex = begin;
}

char Face :: MoveNext( void ){
        char result = 0;

        vertex = vertex->next;

        if( vertex != NULL )
                result = 1;

        return( result );
}

void Face :: MoveLast( void ){
        vertex = end;
}

void Face :: MoveFirstPer( void ){
        in_perspective = begin_per;
}

char Face :: MoveNextPer( void ){
        char result = 0;

        in_perspective = in_perspective->next;

        if( in_perspective != NULL )
                result = 1;

        return( result );
}

void Face :: MoveLastPer( void ){
        in_perspective = end_per;
}


void Face :: AddVertex( float x, float y, float z ){
        Vertex *temp;
        Vertex *temp_per;

        temp = new Vertex;
        temp_per = new Vertex;

        temp->id = cur_vertex_id++;

        temp->orig_x = temp->x  = x;
        temp->orig_y = temp->y  = y;
        temp->orig_z = temp->z  = z;
        temp->orig_w = temp->w  = 0.0;
        temp->normal = NULL;
	temp->texture_normal = NULL;
        temp->next   = NULL;


        temp_per->id = 0;


        temp_per->x  = x;
        temp_per->y  = y;
        temp_per->z  = z;
        temp_per->w  = 0.0;
        temp_per->normal = NULL;
	temp_per->texture_normal = NULL;
        temp_per->next   = NULL;

        if( IsEmpty() ){
                begin = vertex = end = temp;

                begin_per = in_perspective = end_per = temp_per;
        }else{
                end->next = temp;
                end = temp;

                end_per->next = temp_per;
                end_per = temp_per;
        }
}

void Face :: AddVertex( float x, float y, float z, float xn, float yn, float zn ){
        Vertex *temp;
        Vertex *temp_norm;
        Vertex *temp_tex_norm;
        Vertex *temp_per;

        temp = new Vertex;
        temp_norm = new Vertex;
        temp_per = new Vertex;

        temp_norm->id = cur_normal_id++;

        temp_tex_norm->x  = xn;
        temp_tex_norm->y  = yn;
        temp_tex_norm->z  = zn;
        temp_tex_norm->w  = 0.0;
        temp_tex_norm->normal = NULL;
	temp_tex_norm->texture_normal = NULL;
        temp_tex_norm->next   = NULL;

        temp->id = cur_vertex_id++;

        temp->orig_x = temp->x  = x;
        temp->orig_y = temp->y  = y;
        temp->orig_z = temp->z  = z;
        temp->orig_w = temp->w  = 0.0;
        temp->normal = temp_norm;
	temp->texture_normal = temp_tex_norm;
        temp->next   = NULL;

        temp_per->id = 0;

        temp_per->x  = x;
        temp_per->y  = y;
        temp_per->z  = z;
        temp_per->w  = 0.0;
        temp_per->normal = NULL;
        temp_per->next   = NULL;

        if( IsEmpty() ){
                begin = vertex = end = temp;

                begin_per = in_perspective = end_per = temp_per;
        }else{
                end->next = temp;
                end = temp;

                end_per->next = temp_per;
                end_per = temp_per;
        }
}

void Face :: Remove( void ){
        /* remove current node? ... last? */
}

char Face :: IsEmpty( void ){
        return( begin == NULL );
}
                
void Face :: DrawWireFrameNoPerspective( Image *where ){
        Vertex *first;
        Vertex *prev;

        int cols = where->width,
            rows = where->height;

        int rows_offset = 0,
            cols_offset = 0;

        if( rows < cols ){
                cols_offset = (cols-rows)/2;
                cols = rows;
        }
        else if( cols < rows ){
                rows_offset = (rows-cols)/2;
                rows = cols;
        }


        if( IsEmpty() )
                Error( "Cannot draw empty face!\n" );

        MoveFirst();
        first = prev = vertex;
        MoveNext();
        
        while( vertex != NULL ){
                MyDrawLine( cols_offset+(int)(prev->x*cols),rows_offset+(int)((1.0-prev->y)*rows),
cols_offset+(int)(vertex->x*cols),rows_offset+(int)((1.0-vertex->y)*rows), 255, where );
                prev = vertex;
                MoveNext();
        }
        MyDrawLine( cols_offset+(int)(prev->x*cols),rows_offset+(int)((1.0-prev->y)*rows),
cols_offset+(int)(first->x*cols),rows_offset+(int)((1.0-first->y)*rows), 255, where );
}

void Face :: DrawWireFrameInPerspective( Image *where ){
        Vertex *first;
        Vertex *prev;

        int cols = where->width,
            rows = where->height;

        int rows_offset = 0,
            cols_offset = 0;


        if( rows < cols ){
                cols_offset = (cols-rows)/2;
                cols = rows;
        }
        else if( cols < rows ){
                rows_offset = (rows-cols)/2;
                rows = cols;
        }


        if( IsEmpty() )
                Error( "Cannot draw empty face!\n" );

        MoveFirstPer();
        first = prev = in_perspective;
        MoveNextPer();
        
        while( in_perspective != NULL ){
//                printf("(%f %f %f) --> (%f %f %f)\n",prev->x,prev->y,prev->z,in_perspective->x,in_perspective->y,in_perspective->z);fflush(stdout);
//                printf("(%f, %f, %f, %f)\n", cols_offset+(prev->x*cols), rows_offset+((1.0-prev->y)*rows), cols_offset+(in_perspective->x*cols), rows_offset+((1.0-in_perspective->y)*rows) );fflush(stdout);
                MyDrawLine(
cols_offset+(int)(prev->x*cols),rows_offset+(int)((1.0-prev->y)*rows),cols_offset+(int)(in_perspective->x*cols),rows_offset+(int)((1.0-in_perspective->y)*rows),
255, where );
                prev = in_perspective;
                MoveNextPer();
        }
        MyDrawLine( cols_offset+(int)(prev->x*cols),rows_offset+(int)((1.0-prev->y)*rows),
cols_offset+(int)(first->x*cols),rows_offset+(int)((1.0-first->y)*rows), 255, where );
}

/***** functions used for DrawSolid() *****/


/* if ever comparing this to older version of SortScan, remember to add in swaps for ALL variables in older one!! */

void SortScan( Vertex *list ){
        Vertex *head=list;
        float tempx, tempz, tempw;
        Vertex *tempnorm;
	char modified=1;

	while( modified ){
		modified = 0;
		while( list != NULL ){
			if( list->next != NULL ){
				if( list->next->y == list->y ){
					if( list->next->x < list->x ){
                                                tempx = list->x;
                                                tempz = list->z;
						tempw = list->w;
                                                tempnorm = list->normal;

						list->x = list->next->x;
                                                list->z = list->next->z;
						list->w = list->next->w;
                                                list->normal = list->next->normal;

                                                list->next->x = tempx;
                                                list->next->z = tempz;
						list->next->w = tempw;
                                                list->next->normal = tempnorm;
						modified = 1;
					}
				}
			}
			list = list->next;
		}
		list = head;
		break;  //take this out to allow proper drawing of convex polygons
	}
}

float FindHighY( Vertex *vertex ){
        float highest=vertex->y;

        while( vertex != NULL ){
                if( vertex->y > highest )
                        highest = vertex->y;

                vertex = vertex->next;
	}

	return( highest );
}

float FindLowY( Vertex *vertex ){
        float lowest=vertex->y;

        while( vertex != NULL ){
                if( vertex->y < lowest )
                        lowest = vertex->y;

                vertex = vertex->next;
	}

	return( lowest );
}

Vertex *NewVertex( float x, float y, float z ){
        return( NewVertex( x, y, z, 0.0 ) );
}

Vertex *NewVertex( float x, float y, float z, float w ){
        Vertex *temp;

        temp = new Vertex;

        temp->x = x;
        temp->y = y;
        temp->z = z;
        temp->w = w;

        temp->orig_x = x;
        temp->orig_y = y;
        temp->orig_z = z;
        temp->orig_w = w;

	temp->normal = NULL;
        temp->next = NULL;

        return( temp );
}

Vertex *NewVertex( float x, float y, float z, float xn, float yn, float zn ){
        Vertex *temp;
        Vertex *temp_normal;

        temp = new Vertex;
        temp_normal = new Vertex;

        temp_normal->x = xn;
        temp_normal->y = yn;
        temp_normal->z = zn;
        temp_normal->w = 0.0;

        temp_normal->orig_x = x;
        temp_normal->orig_y = y;
        temp_normal->orig_z = z;
        temp_normal->orig_w = 0.0;

        temp_normal->normal = NULL;
        temp_normal->next = NULL;
                   
        temp->x = x;
        temp->y = y;
        temp->z = z;
        temp->w = 0.0;

	temp->orig_x = x;
        temp->orig_y = y;
        temp->orig_z = z;
        temp->orig_w = 0.0;

        temp->normal = temp_normal;
        temp->next = NULL;

        return( temp );
}


void DestroyList( Vertex **vertex ){
        Vertex *temp;

        while( *vertex != NULL ){
                temp = *vertex;
                *vertex = (*vertex)->next;
                delete temp;
        }
}
                                         
Vertex *GetIntersection( Vertex *line1point1, Vertex *line1point2,
                         Vertex *line2point1, Vertex *line2point2 ){
        float u, t;
        float u_numerator, u_denominator;
        float t_numerator, t_denominator;
        float qx=line1point1->x, qy=line1point1->y;
        float rx=line1point2->x, ry=line1point2->y;
        float ax=line2point1->x, ay=line2point1->y;
        float bx=line2point2->x, by=line2point2->y;
        static Vertex *point;

	if( point == NULL )
        	point = NewVertex( 0.0, 0.0, 0.0 );

        u_numerator   = ((qy-ay)*(bx-ax))-((qx-ax)*(by-ay));
        u_denominator = ((qy-ry)*(bx-ax))-((qx-rx)*(by-ay));

        t_numerator   = ((qy-ry)*(qx-ax))-((qx-rx)*(qy-ay));
        t_denominator = ((qy-ry)*(bx-ax))-((qx-rx)*(by-ay));

        if( t_denominator == 0.0 || u_denominator == 0.0 )
                return NULL;
        else{
                u = u_numerator/u_denominator;
                t = t_numerator/t_denominator;

		/*** t is line 2, u is line 1 ***/

                point->x = (float)qx+((float)u*(float)(rx-qx));
                point->y = (float)qy+((float)u*(float)(ry-qy));
                point->w = t;

                if( (t>=0.0 && t<=1.0) && (u>=0.0 && u<=1.0) ){
                        return( point );
                }
                else{
                        return NULL;
                }
        }
}
/*
float FindZ( Vertex *normal, float x, float y, Vertex *vertex ){
        float d;
        float z;

	if( vertex->next != NULL )
		if( vertex->next->next != NULL )
        		vertex = vertex->next->next;

        d = (-1.0*normal->x)*vertex->x - normal->y*vertex->y - normal->z*vertex->z;

	if( normal->z != 0.0 ){
        	z = (float)(x*normal->x + y*normal->y + d)/(float)normal->z;
		return( z );
	}
	else{
		return( 0 );
	}
}


#define FindZ( normal, xx, yy, vertex )  z = (float)(xx*normal->x + yy*normal->y + ((-1.0*normal->x)*vertex->x -
normal->y*vertex->y - normal->z*vertex->z))/(float)normal->z;
 */


/*
**  this function assumes that the screen's height is smaller than (or equal to) its width 
*/
 
void ScanFillTexture( Vertex *vertex, Vertex *normal, int low_color, int high_color, Image *where, Image *from ){
 	int x4=0, y1=0;
	int x3=from->width, y2=0;
	int x2=from->width, y3=from->height;
	int x1=0, y4=from->height;

	Vertex *begin=vertex;
	int left[where->height];		/* leftmost coordinates */
	int right[where->height];		/* rightmost coordinates */

	char linenum_left[where->height];	/* which texture line should be mapped to polygon line */
	char linenum_right[where->height];
	int linenum=1;

	int tx1, ty1;
	int tx2, ty2;
	float dx, dy;

	float gradient;
	float ex, ey;
	float ax;
	float x;
	float cx1, cy1;
	float cx2, cy2;

	float cx3, cy3;
	float cx4, cy4;

	int left_z[where->height];
	int right_z[where->height];
	float cz1, cz2;
	float gradient_z;
	float z;

	int ay, by, y;
	int i, j;	/* stepping through screen coordinates when drawing */
	float tx, ty;	/* stepping through texture coordinates when drawing */

	int x_offset = ((where->width - where->height) >> 1);
	int i_offset;	/* used in for loop for (i*where->width), saves one mult */

	int temp;

	int color;

	memset( left, 0, where->height*sizeof(int) );
	memset( right, 0, where->height*sizeof(int) );

	memset( linenum_left, 0, where->height );
	memset( linenum_right, 0, where->height );

	memset( left_z, 0, where->height*sizeof(int) );	/* z buffer stuff */
	memset( right_z, 0, where->height*sizeof(int) );

	while( vertex->next != NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = vertex->next->x*where->height;
		cy2 = (1.0-vertex->next->y)*where->height;

		cz1 = vertex->z*where->height;
		cz2 = vertex->next->z*where->height;

		if( cy1 != cy2 ){		/* means a horizontal line (doesn't need to be drawn) */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;	

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );

					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );		/* store z to interpolate later */
							linenum_right[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );

					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
							linenum_right[ y ] = linenum;
						}
					}
				}
			}
			else if( cy2 < cy1 ) { 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );

					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );

					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
			}
		}
		vertex = vertex->next;
		++linenum;
	}

	if( vertex->next == NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = begin->x*where->height;
		cy2 = (1.0-begin->y)*where->height;

		cz1 = vertex->z*where->height;
		cz2 = begin->z*where->height;

		if( cy1 != cy2 ){		/* horizontal line */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );

					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
							linenum_right[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );

					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x  );
							right_z[y] = INT( z );
							linenum_right[ y ] = linenum;
						}
					}
				}
			}
			else if( cy2 < cy1 ){ 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );

					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );

					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
			}
		}
		vertex = begin;		/* reset it */
	}

	vertex = begin;

	cx1 = vertex->x*where->height;
	cy1 = (1.0-vertex->y)*where->height;
	cx2 = vertex->next->x*where->height;
	cy2 = (1.0-vertex->next->y)*where->height;
	cx3 = vertex->next->next->x*where->height;
	cy3 = (1.0-vertex->next->next->y)*where->height;
	cx4 = vertex->next->next->next->x*where->height;
	cy4 = (1.0-vertex->next->next->next->y)*where->height;

	for(i=0, i_offset=0; i<where->height; ++i, i_offset+=where->width){
		if( left[i] < right[i] ){

			if( linenum_left[i] == 1 ){
				if( cy1 == cy2 )
					tx1 = x1;
				else
					tx1 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( x2 - x1 ) ) + x1; 

				if( cy1 == cy2 )
					ty1 = y1;
				else
					ty1 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( y2 - y1 ) ) + y1; 
			}
			if( linenum_left[i] == 2 ){
				if( cy2 == cy3 )
					tx1 = x2;
				else
					tx1 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( x3 - x2 ) ) + x2; 

				if( cy2 == cy3 )
					ty1 = y2;
				else
					ty1 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( y3 - y2 ) ) + y2; 
			}
			if( linenum_left[i] == 3 ){
				if( cy3 == cy4 )
					tx1 = x3;
				else
					tx1 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( x4 - x3 ) ) + x3; 
				
				if( cy3 == cy4 )
					ty1 = y3;
				else
					ty1 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( y4 - y3 ) ) + y3; 
			}
			if( linenum_left[i] == 4 ){
				if( cy4 == cy1 )
					tx1 = x4;
				else
					tx1 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( x1 - x4 ) ) + x4; 

				if( cy4 == cy1 )
					ty1 = y4;
				else
					ty1 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( y1 - y4 ) ) + y4; 
			}

			if( linenum_right[i] == 1 ){
				if( cy1 == cy2 )
					tx2 = x1;
				else
					tx2 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( x2 - x1 ) ) + x1; 

				if( cy1 == cy2 )
					ty2 = y1;
				else
					ty2 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( y2 - y1 ) ) + y1; 
			}
			if( linenum_right[i] == 2 ){
				if( cy2 == cy3 )
					tx2 = x2;
				else
					tx2 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( x3 - x2 ) ) + x2; 

				if( cy2 == cy3 )
					ty2 = y2;
				else
					ty2 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( y3 - y2 ) ) + y2; 
			}
			if( linenum_right[i] == 3 ){
				if( cy3 == cy4 )
					tx2 = x3;
				else
					tx2 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( x4 - x3 ) ) + x3; 
				
				if( cy3 == cy4 )
					ty2 = y3;
				else
					ty2 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( y4 - y3 ) ) + y3; 
			}
			if( linenum_right[i] == 4 ){
				if( cy4 == cy1 )
					tx2 = x4;
				else
					tx2 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( x1 - x4 ) ) + x4; 

				if( cy4 == cy1 )
					ty2 = y4;
				else
					ty2 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( y1 - y4 ) ) + y4; 
			}

			if( tx1 < 0 ) tx1 = 0;
			if( tx2 < 0 ) tx2 = 0;
			if( ty1 < 0 ) ty1 = 0;
			if( ty2 < 0 ) ty2 = 0;

			if( tx1 > 255 ) tx1 = 255;
			if( tx2 > 255 ) tx2 = 255;
			if( ty1 > 255 ) ty1 = 255;
			if( ty2 > 255 ) ty2 = 255;

			dx = (float)( tx2 - tx1 ) / (float)( right[i] - left[i] );
			dy = (float)( ty2 - ty1 ) / (float)( right[i] - left[i] );

			gradient_z = (float)(right_z[i] - left_z[i]) / (float)(right[i] - left[i]);	/* to interpolate z */
			for(j=left[i]+x_offset, tx = tx1, ty = ty1, z=left_z[i]; j<=(right[i]+x_offset); ++j, tx += dx, ty += dy, z+=gradient_z  ){
				temp = i_offset + j;
				if( InBounds( temp, (where->width*where->height) ) )
				if( InBounds( j, where->width ) )
				if( INT(z) < 0 )
				if( INT(z) > z_buffer[ temp ] ){
					SetPixel( j, i, from->bitmap[ ( ((int)tx) * from->width ) + (int)ty ], where );
					z_buffer[ temp ] = INT(z);
				}
			}
		}
	}
}

void ScanFillFastPhong( Vertex *vertex, Vertex *normal, int low_color, int high_color, Image *where, char *from, int x1, int y1,
int x2,
int y2, int x3, int y3, int x4, int y4 ){
	Vertex *begin=vertex;
	int left[where->height];		/* leftmost coordinates */
	int right[where->height];		/* rightmost coordinates */

	char linenum_left[where->height];	/* which texture line should be mapped to polygon line */
	char linenum_right[where->height];
	int linenum=1;

	int tx1, ty1;
	int tx2, ty2;
	float dx, dy;

	float gradient;
	float ex, ey;
	float ax;
	float x;
	float cx1, cy1;
	float cx2, cy2;

	float cx3, cy3;
	float cx4, cy4;

	int left_z[where->height];
	int right_z[where->height];
	float cz1, cz2;
	float gradient_z;
	float z;

	int ay, by, y;
	int i, j;	/* stepping through screen coordinates when drawing */
	float tx, ty;	/* stepping through texture coordinates when drawing */

	int x_offset = ((where->width - where->height) >> 1);
	int i_offset;	/* used in for loop for (i*where->width), saves one mult */

	int temp;

	int color;

	memset( left, 0, where->height*sizeof(int) );
	memset( right, 0, where->height*sizeof(int) );

	memset( linenum_left, 0, where->height );
	memset( linenum_right, 0, where->height );

	memset( left_z, 0, where->height*sizeof(int) );	/* z buffer stuff */
	memset( right_z, 0, where->height*sizeof(int) );

	while( vertex->next != NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = vertex->next->x*where->height;
		cy2 = (1.0-vertex->next->y)*where->height;
		
		cz1 = vertex->z*where->height;
		cz2 = vertex->next->z*where->height;

		if( cy1 != cy2 ){		/* means a horizontal line (doesn't need to be drawn) */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;	

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );		/* store z to interpolate later */
							linenum_right[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
							linenum_right[ y ] = linenum;
						}
					}
				}
			}
			else if( cy2 < cy1 ) { 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
			}
		}
		vertex = vertex->next;
		++linenum;
	}

	if( vertex->next == NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = begin->x*where->height;
		cy2 = (1.0-begin->y)*where->height;

		cz1 = vertex->z*where->height;
		cz2 = begin->z*where->height;

		if( cy1 != cy2 ){		/* horizontal line */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
							linenum_right[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x  );
							right_z[y] = INT( z );
							linenum_right[ y ] = linenum;
						}
					}
				}
			}
			else if( cy2 < cy1 ){ 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
			}
		}
		vertex = begin;		/* reset it */
	}

	cx1 = vertex->x*where->height;
	cy1 = (1.0-vertex->y)*where->height;
	cx2 = vertex->next->x*where->height;
	cy2 = (1.0-vertex->next->y)*where->height;
	cx3 = vertex->next->next->x*where->height;
	cy3 = (1.0-vertex->next->next->y)*where->height;
	cx4 = vertex->next->next->next->x*where->height;
	cy4 = (1.0-vertex->next->next->next->y)*where->height;


	for(i=0, i_offset=0; i<where->height; ++i, i_offset+=where->width){
		if( left[i] < right[i] ){

			if( linenum_left[i] == 1 ){
				if( cx1 == cx2 )
					tx1 = x1;
				else
					tx1 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( x2 - x1 ) ) + x1; 

				if( cy1 == cy2 )
					ty1 = y1;
				else
					ty1 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( y2 - y1 ) ) + y1; 
			}
			if( linenum_left[i] == 2 ){
				if( cx2 == cx3 )
					tx1 = x2;
				else
					tx1 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( x3 - x2 ) ) + x2; 

				if( cy2 == cy3 )
					ty1 = y2;
				else
					ty1 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( y3 - y2 ) ) + y2; 
			}
			if( linenum_left[i] == 3 ){
				if( cx3 == cx4 )
					tx1 = x3;
				else
					tx1 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( x4 - x3 ) ) + x3; 
				
				if( cy3 == cy4 )
					ty1 = y3;
				else
					ty1 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( y4 - y3 ) ) + y3; 
			}
			if( linenum_left[i] == 4 ){
				if( cx4 == cx1 )
					tx1 = x4;
				else
					tx1 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( x1 - x4 ) ) + x4; 

				if( cy4 == cy1 )
					ty1 = y4;
				else
					ty1 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( y1 - y4 ) ) + y4; 
			}

			if( linenum_right[i] == 1 ){
				if( cx1 == cx2 )
					tx2 = x1;
				else
					tx2 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( x2 - x1 ) ) + x1; 

				if( cy1 == cy2 )
					ty2 = y1;
				else
					ty2 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( y2 - y1 ) ) + y1; 
			}
			if( linenum_right[i] == 2 ){
				if( cx2 == cx3 )
					tx2 = x2;
				else
					tx2 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( x3 - x2 ) ) + x2; 

				if( cy2 == cy3 )
					ty2 = y2;
				else
					ty2 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( y3 - y2 ) ) + y2; 
			}
			if( linenum_right[i] == 3 ){
				if( cx3 == cx4 )
					tx2 = x3;
				else
					tx2 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( x4 - x3 ) ) + x3; 
				
				if( cy3 == cy4 )
					ty2 = y3;
				else
					ty2 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( y4 - y3 ) ) + y3; 
			}
			if( linenum_right[i] == 4 ){
				if( cx4 == cx1 )
					tx2 = x4;
				else
					tx2 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( x1 - x4 ) ) + x4; 

				if( cy4 == cy1 )
					ty2 = y4;
				else
					ty2 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( y1 - y4 ) ) + y4; 
			}

			if( tx1 < 0 ) tx1 = 0;
			if( tx2 < 0 ) tx2 = 0;
			if( ty1 < 0 ) ty1 = 0;
			if( ty2 < 0 ) ty2 = 0;

			if( tx1 > 255 ) tx1 = 255;
			if( tx2 > 255 ) tx2 = 255;
			if( ty1 > 255 ) ty1 = 255;
			if( ty2 > 255 ) ty2 = 255;

			dx = (float)( tx2 - tx1 ) / (float)( right[i] - left[i] );
			dy = (float)( ty2 - ty1 ) / (float)( right[i] - left[i] );

			gradient_z = (float)(right_z[i] - left_z[i]) / (float)(right[i] - left[i]);	/* to interpolate z */

			for(j=left[i]+x_offset, tx = tx1, ty = ty1, z=left_z[i]; j<=(right[i]+x_offset); ++j, tx += dx, ty += dy, z+=gradient_z  ){
				temp = i_offset + j;
				if( InBounds( temp, (where->width*where->height) ) )
				if( InBounds( j, where->width ) )
				if( INT(z) < 0 )
				if( INT(z) > z_buffer[ temp ] ){
					SetPixel( j, i, from[ ( ((int)tx) << 8 ) + (int)ty ], where );
					z_buffer[ temp ] = INT( z );
				}
			}
		}
	}
}
void ScanFillStretchedTexture( Vertex *vertex, Vertex *normal, int low_color, int high_color, Image *where, Image *from, int x1, int y1,
int x2,
int y2, int x3, int y3, int x4, int y4 ){
	Vertex *begin=vertex;
	int left[where->height];		/* leftmost coordinates */
	int right[where->height];		/* rightmost coordinates */

	char linenum_left[where->height];	/* which texture line should be mapped to polygon line */
	char linenum_right[where->height];
	int linenum=1;

	int tx1, ty1;
	int tx2, ty2;
	float dx, dy;

	float gradient;
	float ex, ey;
	float ax;
	float x;
	float cx1, cy1;
	float cx2, cy2;

	float cx3, cy3;
	float cx4, cy4;

	int left_z[where->height];
	int right_z[where->height];
	float cz1, cz2;
	float gradient_z;
	float z;

	int ay, by, y;
	int i, j;	/* stepping through screen coordinates when drawing */
	float tx, ty;	/* stepping through texture coordinates when drawing */

	int x_offset = ((where->width - where->height) >> 1);
	int i_offset;	/* used in for loop for (i*where->width), saves one mult */

	int temp;

	int color;

	memset( left, 0, where->height*sizeof(int) );
	memset( right, 0, where->height*sizeof(int) );

	memset( linenum_left, 0, where->height );
	memset( linenum_right, 0, where->height );

	memset( left_z, 0, where->height*sizeof(int) );	/* z buffer stuff */
	memset( right_z, 0, where->height*sizeof(int) );

	while( vertex->next != NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = vertex->next->x*where->height;
		cy2 = (1.0-vertex->next->y)*where->height;
		
		cz1 = vertex->z*where->height;
		cz2 = vertex->next->z*where->height;

		if( cy1 != cy2 ){		/* means a horizontal line (doesn't need to be drawn) */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;	

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );		/* store z to interpolate later */
							linenum_right[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
							linenum_right[ y ] = linenum;
						}
					}
				}
			}
			else if( cy2 < cy1 ) { 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
			}
		}
		vertex = vertex->next;
		++linenum;
	}

	if( vertex->next == NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = begin->x*where->height;
		cy2 = (1.0-begin->y)*where->height;

		cz1 = vertex->z*where->height;
		cz2 = begin->z*where->height;

		if( cy1 != cy2 ){		/* horizontal line */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
							linenum_right[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x  );
							right_z[y] = INT( z );
							linenum_right[ y ] = linenum;
						}
					}
				}
			}
			else if( cy2 < cy1 ){ 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							linenum_left[ y ] = linenum;
						}
					}
				}
			}
		}
		vertex = begin;		/* reset it */
	}

	cx1 = vertex->x*where->height;
	cy1 = (1.0-vertex->y)*where->height;
	cx2 = vertex->next->x*where->height;
	cy2 = (1.0-vertex->next->y)*where->height;
	cx3 = vertex->next->next->x*where->height;
	cy3 = (1.0-vertex->next->next->y)*where->height;
	cx4 = vertex->next->next->next->x*where->height;
	cy4 = (1.0-vertex->next->next->next->y)*where->height;


	for(i=0, i_offset=0; i<where->height; ++i, i_offset+=where->width){
		if( left[i] < right[i] ){

			if( linenum_left[i] == 1 ){
				if( cx1 == cx2 )
					tx1 = x1;
				else
					tx1 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( x2 - x1 ) ) + x1; 

				if( cy1 == cy2 )
					ty1 = y1;
				else
					ty1 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( y2 - y1 ) ) + y1; 
			}
			if( linenum_left[i] == 2 ){
				if( cx2 == cx3 )
					tx1 = x2;
				else
					tx1 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( x3 - x2 ) ) + x2; 

				if( cy2 == cy3 )
					ty1 = y2;
				else
					ty1 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( y3 - y2 ) ) + y2; 
			}
			if( linenum_left[i] == 3 ){
				if( cx3 == cx4 )
					tx1 = x3;
				else
					tx1 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( x4 - x3 ) ) + x3; 
				
				if( cy3 == cy4 )
					ty1 = y3;
				else
					ty1 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( y4 - y3 ) ) + y3; 
			}
			if( linenum_left[i] == 4 ){
				if( cx4 == cx1 )
					tx1 = x4;
				else
					tx1 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( x1 - x4 ) ) + x4; 

				if( cy4 == cy1 )
					ty1 = y4;
				else
					ty1 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( y1 - y4 ) ) + y4; 
			}

			if( linenum_right[i] == 1 ){
				if( cx1 == cx2 )
					tx2 = x1;
				else
					tx2 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( x2 - x1 ) ) + x1; 

				if( cy1 == cy2 )
					ty2 = y1;
				else
					ty2 = (int)( ( ( i - cy1 ) / ( cy2 - cy1 ) ) * ( y2 - y1 ) ) + y1; 
			}
			if( linenum_right[i] == 2 ){
				if( cx2 == cx3 )
					tx2 = x2;
				else
					tx2 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( x3 - x2 ) ) + x2; 

				if( cy2 == cy3 )
					ty2 = y2;
				else
					ty2 = (int)( ( ( i - cy2 ) / ( cy3 - cy2 ) ) * ( y3 - y2 ) ) + y2; 
			}
			if( linenum_right[i] == 3 ){
				if( cx3 == cx4 )
					tx2 = x3;
				else
					tx2 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( x4 - x3 ) ) + x3; 
				
				if( cy3 == cy4 )
					ty2 = y3;
				else
					ty2 = (int)( ( ( i - cy3 ) / ( cy4 - cy3 ) ) * ( y4 - y3 ) ) + y3; 
			}
			if( linenum_right[i] == 4 ){
				if( cx4 == cx1 )
					tx2 = x4;
				else
					tx2 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( x1 - x4 ) ) + x4; 

				if( cy4 == cy1 )
					ty2 = y4;
				else
					ty2 = (int)( ( ( i - cy4 ) / ( cy1 - cy4 ) ) * ( y1 - y4 ) ) + y4; 
			}

			if( tx1 < 0 ) tx1 = 0;
			if( tx2 < 0 ) tx2 = 0;
			if( ty1 < 0 ) ty1 = 0;
			if( ty2 < 0 ) ty2 = 0;

			if( tx1 > from->width ) tx1 = from->width;
			if( tx2 > from->width ) tx2 = from->width;
			if( ty1 > from->height ) ty1 = from->height;
			if( ty2 > from->height ) ty2 = from->height;

			dx = (float)( tx2 - tx1 ) / (float)( right[i] - left[i] );
			dy = (float)( ty2 - ty1 ) / (float)( right[i] - left[i] );

			gradient_z = (float)(right_z[i] - left_z[i]) / (float)(right[i] - left[i]);	/* to interpolate z */

			for(j=left[i]+x_offset, tx = tx1, ty = ty1, z=left_z[i]; j<=(right[i]+x_offset); ++j, tx += dx, ty += dy, z+=gradient_z  ){
				temp = i_offset + j;
				if( InBounds( temp, (where->width*where->height) ) )
				if( InBounds( j, where->width ) )
				if( INT(z) < 0 )
				if( INT(z) > z_buffer[ temp ] ){
					SetPixel( j, i, from->bitmap[ ( ((int)tx) * from->width ) + (int)ty ], where );
					z_buffer[ temp ] = INT( z );
				}
			}
		}
	}
}

void ScanFillTransparent( Vertex *vertex, Vertex *normal, int color, int low_color, int high_color, Image *where ){
	Vertex *begin=vertex;
	int left[where->height];		/* leftmost coordinates */
	int right[where->height];		/* rightmost coordinates */

	float gradient;
	float ex, ey;
	float ax;
	float x;
	float cx1, cy1;
	float cx2, cy2;

	float cx3, cy3;
	float cx4, cy4;

	int left_z[where->height];
	int right_z[where->height];
	float cz1, cz2;
	float gradient_z;
	float z;

	int ay, by, y;
	int i, j;	/* stepping through screen coordinates when drawing */

	int x_offset = ((where->width - where->height) >> 1);
	int i_offset;	/* used in for loop for (i*where->width), saves one mult */

	int temp;
	unsigned char oldcolor, newcolor;

	memset( left, 0, where->height*sizeof(int) );
	memset( right, 0, where->height*sizeof(int) );

	memset( left_z, 0, where->height*sizeof(int) );	/* z buffer stuff */
	memset( right_z, 0, where->height*sizeof(int) );

	while( vertex->next != NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = vertex->next->x*where->height;
		cy2 = (1.0-vertex->next->y)*where->height;
		
		cz1 = vertex->z*where->height;
		cz2 = vertex->next->z*where->height;

		if( cy1 != cy2 ){		/* means a horizontal line (doesn't need to be drawn) */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;	

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
						}
					}
				}
			}
			else if( cy2 < cy1 ) { 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
						}
					}
				}
			}
		}
		vertex = vertex->next;
	}

	if( vertex->next == NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = begin->x*where->height;
		cy2 = (1.0-begin->y)*where->height;

		cz1 = vertex->z*where->height;
		cz2 = begin->z*where->height;

		if( cy1 != cy2 ){		/* horizontal line */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x  );
						}
					}
				}
			}
			else if( cy2 < cy1 ){ 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
						}
					}
				}
			}
		}
		vertex = begin;		/* reset it */
	}

	for(i=0, i_offset=0; i<where->height; ++i, i_offset+=where->width){
		if( left[i] < right[i] ){
			for(j=left[i]+x_offset; j<=(right[i]+x_offset); ++j ){
				if( InBounds( j, where->width ) ){
					oldcolor = (unsigned char)((((unsigned char)where->bitmap[ i_offset + j ])/255.0)*(high_color-low_color)) + low_color;
					newcolor = (unsigned char)(((high_color-oldcolor) * 0.7) + ((high_color-color) * 0.3)) + low_color;
					if( newcolor < low_color ) newcolor = low_color;
					if( newcolor > high_color ) newcolor = high_color;
				if( InBounds( j, where->width ) )

					SetPixel( j, i, newcolor, where );
				}
			}
		}
		else if( left[i] > right[i] ){
			for(j=right[i]+x_offset; j<=(left[i]+x_offset); ++j ){
				if( InBounds( j, where->width ) ){
					oldcolor = (unsigned char)((((unsigned char)where->bitmap[ i_offset + j ])/255.0)*(high_color-low_color)) + low_color;
					newcolor = (unsigned char)(((high_color-oldcolor) * 0.7) + ((high_color-color) * 0.3)) + low_color;
					if( newcolor < low_color ) newcolor = low_color;
					if( newcolor > high_color ) newcolor = high_color;
				if( InBounds( j, where->width ) )

					SetPixel( j, i, newcolor, where );
				}
			}
		}
	}
}

void ScanFill( Vertex *vertex, Vertex *normal, int color, Image *where ){
	Vertex *begin=vertex;
	int left[where->height];		/* leftmost coordinates */
	int right[where->height];		/* rightmost coordinates */

	float gradient;
	float ex, ey;
	float ax;
	float x;
	float cx1, cy1;
	float cx2, cy2;

	float cx3, cy3;
	float cx4, cy4;

	int left_z[where->height];
	int right_z[where->height];
	float cz1, cz2;
	float gradient_z;
	float z;

	int ay, by, y;
	int i, j;	/* stepping through screen coordinates when drawing */

	int x_offset = ((where->width - where->height) >> 1);
	int i_offset;	/* used in for loop for (i*where->width), saves one mult */

	int temp;

	memset( left, 0, where->height*sizeof(int) );
	memset( right, 0, where->height*sizeof(int) );

	memset( left_z, 0, where->height*sizeof(int) );	/* z buffer stuff */
	memset( right_z, 0, where->height*sizeof(int) );

	while( vertex->next != NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = vertex->next->x*where->height;
		cy2 = (1.0-vertex->next->y)*where->height;
		
		cz1 = vertex->z*where->height;
		cz2 = vertex->next->z*where->height;

		if( cy1 != cy2 ){		/* means a horizontal line (doesn't need to be drawn) */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;	

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );		/* store z to interpolate later */
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
						}
					}
				}
			}
			else if( cy2 < cy1 ) { 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
						}
					}
				}
			}
		}
		vertex = vertex->next;
	}

	if( vertex->next == NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = begin->x*where->height;
		cy2 = (1.0-begin->y)*where->height;

		cz1 = vertex->z*where->height;
		cz2 = begin->z*where->height;

		if( cy1 != cy2 ){		/* horizontal line */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x+=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1; y<=by; x-=gradient, y++, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x  );
							right_z[y] = INT( z );
						}
					}
				}
			}
			else if( cy2 < cy1 ){ 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x+=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
						}
					}
				}
				else{
					ax = cx1 - ex;
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1; y>=by; x-=gradient, y--, z+=gradient_z ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
						}
					}
				}
			}
		}
		vertex = begin;		/* reset it */
	}

	for(i=0, i_offset=0; i<where->height; ++i, i_offset+=where->width){
		if( left[i] < right[i] ){
			gradient_z = (float)(right_z[i] - left_z[i]) / (float)(right[i] - left[i]);	/* to interpolate z */

			for(j=left[i]+x_offset, z=left_z[i]; j<=(right[i]+x_offset); ++j, z+=gradient_z  ){
				temp = i_offset + j;
				if( InBounds( temp, (where->width*where->height) ) )
				if( InBounds( j, where->width ) )
				if( INT(z) < 0 )
				if( INT(z) > z_buffer[ temp ] ){
					SetPixel( j, i, color, where );
					z_buffer[ temp ] = INT( z );
				}
			}
		}
	}
}

void ScanFillGouraud( Vertex *vertex, Vertex *normal, int low_color, int high_color, Image *where ){
	Vertex *begin=vertex;
	int left[where->height];		/* leftmost coordinates */
	int right[where->height];		/* rightmost coordinates */

	float gradient;
	float ex, ey;
	float ax;
	float x;
	float cx1, cy1;
	float cx2, cy2;

	float cx3, cy3;
	float cx4, cy4;

	float left_nl[where->height], right_nl[where->height];
	float cnl1, cnl2;
	float gradient_nl;
	float nl; 

	int left_z[where->height];
	int right_z[where->height];
	float cz1, cz2;
	float gradient_z;
	float z;

	int ay, by, y;
	int i, j;	/* stepping through screen coordinates when drawing */

	int x_offset = ((where->width - where->height) >> 1);
	int i_offset;	/* used in for loop for (i*where->width), saves one mult */

	int ncolor;
	int temp;
	int hi_sub_lo = high_color - low_color;

	memset( left, 0, where->height*sizeof(int) );
	memset( right, 0, where->height*sizeof(int) );

	memset( left_z, 0, where->height*sizeof(int) );	/* z buffer stuff */
	memset( right_z, 0, where->height*sizeof(int) );

	memset( left_nl, 0, where->height*sizeof(float) );	/* gouraud info */
	memset( right_nl, 0, where->height*sizeof(float) );

	while( vertex->next != NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = vertex->next->x*where->height;
		cy2 = (1.0-vertex->next->y)*where->height;
		
		cz1 = vertex->z*where->height;
		cz2 = vertex->next->z*where->height;

		cnl1 = DotProduct( vertex->normal, light );
		cnl2 = DotProduct( vertex->next->normal, light );

		if( INT( cy1 ) != INT( cy2 ) ){		/* means a horizontal line (doesn't need to be drawn) */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0.0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			gradient_nl = (cnl2 - cnl1) / fabs(cy2 - cy1);		/* for gouraud */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;	

				if( cx1 < cx2 ){
					ax = cx1 + ex;
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1, nl=cnl1; y<=by; x+=gradient, y++, z+=gradient_z, nl+=gradient_nl ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );		/* store z to interpolate later */
							right_nl[y] = nl;
						}
					}
				}
				else{
					ax = INT( cx1 - ex );
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1, nl=cnl1; y<=by; x-=gradient, y++, z+=gradient_z, nl+=gradient_nl ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
							right_nl[y] = nl;
						}
					}
				}
			}
			else if( cy2 < cy1 ) { 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = INT( cx1 + ex );
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1, nl=cnl1; y>=by; x+=gradient, y--, z+=gradient_z, nl+=gradient_nl ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							left_nl[y] = nl;
						}
					}
				}
				else{
					ax = INT( cx1 - ex );
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1, nl=cnl1; y>=by; x-=gradient, y--, z+=gradient_z, nl+=gradient_nl ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							left_nl[y] = nl;
						}
					}
				}
			}
		}
		vertex = vertex->next;
	}

	if( vertex->next == NULL ){
		cx1 = vertex->x*where->height;
		cy1 = (1.0-vertex->y)*where->height;
		cx2 = begin->x*where->height;
		cy2 = (1.0-begin->y)*where->height;

		cz1 = vertex->z*where->height;
		cz2 = begin->z*where->height;

		cnl1 = DotProduct( vertex->normal, light );
		cnl2 = DotProduct( begin->normal, light );

		if( INT( cy1 ) != INT( cy2 ) ){		/* horizontal line */
			gradient = (cx2 - cx1) / (cy2 - cy1);
			if( gradient < 0.0 ) gradient *= -1.0;

			gradient_z = (cz2 - cz1) / fabs(cy2 - cy1);		/* for z buffer */

			gradient_nl = (cnl2 - cnl1) / fabs(cy2 - cy1);		/* for gouraud */

			if( cy1 < cy2 ){	/* on the left */
				ey = INT( cy1 + 1.0 ) - cy1;
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = INT( cx1 + ex );
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1, nl=cnl1; y<=by; x+=gradient, y++, z+=gradient_z, nl+=gradient_nl ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x );
							right_z[y] = INT( z );
							right_nl[y] = nl;
						}
					}
				}
				else{
					ax = INT( cx1 - ex );
					ay = INT( cy1 + 1.0 );

					by = INT( cy2 );
					for(x=ax, y=ay, z=cz1, nl=cnl1; y<=by; x-=gradient, y++, z+=gradient_z, nl+=gradient_nl ){
						if( InBounds( y, where->height ) ){
							right[y] = INT( x  );
							right_z[y] = INT( z );
							right_nl[y] = nl;
						}
					}
				}
			}
			else if( cy2 < cy1 ){ 
				ey = cy1 - INT( cy1 );
				ex = gradient*ey;

				if( cx1 < cx2 ){
					ax = INT( cx1 + ex );
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1, nl=cnl1; y>=by; x+=gradient, y--, z+=gradient_z, nl+=gradient_nl ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							left_nl[y] = nl;
						}
					}
				}
				else{
					ax = INT( cx1 - ex );
					ay = INT( cy1 );

					by = INT( cy2 + 1.0 );
					for(x=ax, y=ay, z=cz1, nl=cnl1; y>=by; x-=gradient, y--, z+=gradient_z, nl+=gradient_nl ){
						if( InBounds( y, where->height ) ){
							left[y] = INT( x );
							left_z[y] = INT( z );
							left_nl[y] = nl;
						}
					}
				}
			}
		}
		vertex = begin;		/* reset it */
	}

	for(i=0, i_offset=0; i<where->height; ++i, i_offset+=where->width){
		if( left[i] < right[i] ){
			gradient_z = (float)(right_z[i] - left_z[i]) / (float)(right[i] - left[i]);	/* to interpolate z */
			gradient_nl = (float)(right_nl[i] - left_nl[i]) / (float)(right[i] - left[i]);	/* to interpolate z */

			for(j=left[i]+x_offset, z=left_z[i], nl=left_nl[i]; j<=(right[i]+x_offset); ++j, z+=gradient_z, nl+=gradient_nl ){
				temp = i_offset + j;
				if( InBounds( temp, (where->width*where->height) ) )
				if( InBounds( j, where->width ) )
				if( INT( z ) < 0 )
				if( INT(z) > z_buffer[ temp ] ){
					ncolor = (int)(nl*hi_sub_lo + low_color);
					if( ncolor < low_color ) ncolor = low_color;
					if( ncolor > high_color ) ncolor = high_color;
					SetPixel( j, i, ncolor, where );
					z_buffer[ temp ] = INT( z );
				}
			}
		}
	}
}

/*****************************/

float DotProduct( Vertex *vector1, Vertex *vector2 ){
        float dot;

	if( vector1 != NULL && vector2 != NULL ){
		dot = (vector1->x*vector2->x) + (vector1->y*vector2->y) + (vector1->z*vector2->z);
	}
	else{
		dot = .999;
	}

	return( dot );
}

void Face :: DrawSolidTexture( Image *where ){
        if( use_perspective == 0.0 ){
		MoveFirst();
	        ScanFillTexture( vertex, normal_to_face, low_color, high_color, where, &texture );
	}
	else{
		MoveFirstPer();
        	ScanFillTexture( in_perspective, normal_to_face, low_color, high_color, where, &texture );
	}
}

void Face :: DrawSolidStretchedTexture( Image *where ){
        Vertex *temp;
        float xn=0.0, yn=0.0, zn=0.0;
        int count=0;
        float px1, py1,
               px2, py2,
               px3, py3,
               px4, py4;

	int x1, y1,
	    x2, y2,
	    x3, y3,
	    x4, y4;

        if( use_perspective == 0.0 ){
                MoveFirst();
                
                temp = begin;  
                while( temp != NULL ){
                        if( count == 0 ){
                                px1 = temp->texture_normal->x;
                                py1 = temp->texture_normal->y;
                        }
                        if( count == 1 ){
                                px2 = temp->texture_normal->x;
                                py2 = temp->texture_normal->y;
                        }
                        if( count == 2 ){
                                px3 = temp->texture_normal->x;
                                py3 = temp->texture_normal->y;
                        }
                        if( count == 3 ){
                                px4 = temp->texture_normal->x;
                                py4 = temp->texture_normal->y;
                        }   
                        ++count;
                        temp = temp->next;
                }

		MoveFirst();

		x1 = (int)((px1*128)+127);
		y1 = (int)((py1*128)+127);
		x2 = (int)((px2*128)+127);
		y2 = (int)((py2*128)+127);
		x3 = (int)((px3*128)+127);
		y3 = (int)((py3*128)+127);
		x4 = (int)((px4*128)+127);
		y4 = (int)((py4*128)+127);

/*
		if( x1 < 0 ) x1 = 0;
		if( x2 < 0 ) x2 = 0;
		if( x3 < 0 ) x3 = 0;
		if( x4 < 0 ) x4 = 0;

		if( x1 > 255 ) x1 = 255;
		if( x2 > 255 ) x2 = 255;
		if( x3 > 255 ) x3 = 255;
		if( x4 > 255 ) x4 = 255;

		if( y1 < 0 ) y1 = 0;
		if( y2 < 0 ) y2 = 0;
		if( y3 < 0 ) y3 = 0;
		if( y4 < 0 ) y4 = 0;

		if( y1 > 255 ) y1 = 255;
		if( y2 > 255 ) y2 = 255;
		if( y3 > 255 ) y3 = 255;
		if( y4 > 255 ) y4 = 255;
*/

        	ScanFillStretchedTexture( vertex, normal_to_face, low_color, high_color, where, &texture,
			x1, y1,
			x2, y2,
			x3, y3,
			x4, y4 );
            
        }else{
                MoveFirstPer();

                temp = begin_per;  
                while( temp != NULL ){
                        if( count == 0 ){
                                px1 = temp->texture_normal->x;
                                py1 = temp->texture_normal->y;
                        }
                        if( count == 1 ){
                                px2 = temp->texture_normal->x;
                                py2 = temp->texture_normal->y;
                        }
                        if( count == 2 ){
                                px3 = temp->texture_normal->x;
                                py3 = temp->texture_normal->y;
                        }
                        if( count == 3 ){
                                px4 = temp->texture_normal->x;
                                py4 = temp->texture_normal->y;
                        }   
                        ++count;
                        temp = temp->next;
                }

                MoveFirstPer();

		x1 = (int)((px1*128)+127);
		y1 = (int)((py1*128)+127);
		x2 = (int)((px2*128)+127);
		y2 = (int)((py2*128)+127);
		x3 = (int)((px3*128)+127);
		y3 = (int)((py3*128)+127);
		x4 = (int)((px4*128)+127);
		y4 = (int)((py4*128)+127);

/*
		if( x1 < 0 ) x1 = 0;
		if( x2 < 0 ) x2 = 0;
		if( x3 < 0 ) x3 = 0;
		if( x4 < 0 ) x4 = 0;

		if( x1 > 255 ) x1 = 255;
		if( x2 > 255 ) x2 = 255;
		if( x3 > 255 ) x3 = 255;
		if( x4 > 255 ) x4 = 255;

		if( y1 < 0 ) y1 = 0;
		if( y2 < 0 ) y2 = 0;
		if( y3 < 0 ) y3 = 0;
		if( y4 < 0 ) y4 = 0;

		if( y1 > 255 ) y1 = 255;
		if( y2 > 255 ) y2 = 255;
		if( y3 > 255 ) y3 = 255;
		if( y4 > 255 ) y4 = 255;
*/                          
        	ScanFillStretchedTexture( in_perspective, normal_to_face, low_color, high_color, where, &texture,
			x1, y1,
			x2, y2,
			x3, y3,
			x4, y4 );
        }

}


void Face :: DrawSolidNoShade( Image *where ){
        if( use_perspective == 0.0 ){
		MoveFirst();
	        ScanFill( vertex, normal_to_face, high_color, where );
	}
	else{
		MoveFirstPer();
        	ScanFill( in_perspective, normal_to_face, high_color, where );
	}
}

void Face :: DrawSolidFastPhongShade( Image *where, char *phongmap ){
        Vertex *temp;
        float xn=0.0, yn=0.0, zn=0.0;
        int count=0;
        float px1, py1,
               px2, py2,
               px3, py3,
               px4, py4;

	int x1, y1,
	    x2, y2,
	    x3, y3,
	    x4, y4;

        if( use_perspective == 0.0 ){
                MoveFirst();
                
                temp = begin;  
                while( temp != NULL ){
                        if( count == 0 ){
                                px1 = temp->normal->x;
                                py1 = temp->normal->y;
                        }
                        if( count == 1 ){
                                px2 = temp->normal->x;
                                py2 = temp->normal->y;
                        }
                        if( count == 2 ){
                                px3 = temp->normal->x;
                                py3 = temp->normal->y;
                        }
                        if( count == 3 ){
                                px4 = temp->normal->x;
                                py4 = temp->normal->y;
                        }   
                        ++count;
                        temp = temp->next;
                }

		MoveFirst();

		x1 = (int)((px1*128)+127);
		y1 = (int)((py1*128)+127);
		x2 = (int)((px2*128)+127);
		y2 = (int)((py2*128)+127);
		x3 = (int)((px3*128)+127);
		y3 = (int)((py3*128)+127);
		x4 = (int)((px4*128)+127);
		y4 = (int)((py4*128)+127);

		if( x1 < 0 ) x1 = 0;
		if( x2 < 0 ) x2 = 0;
		if( x3 < 0 ) x3 = 0;
		if( x4 < 0 ) x4 = 0;

		if( x1 > 255 ) x1 = 255;
		if( x2 > 255 ) x2 = 255;
		if( x3 > 255 ) x3 = 255;
		if( x4 > 255 ) x4 = 255;

		if( y1 < 0 ) y1 = 0;
		if( y2 < 0 ) y2 = 0;
		if( y3 < 0 ) y3 = 0;
		if( y4 < 0 ) y4 = 0;

		if( y1 > 255 ) y1 = 255;
		if( y2 > 255 ) y2 = 255;
		if( y3 > 255 ) y3 = 255;
		if( y4 > 255 ) y4 = 255;

        	ScanFillFastPhong( vertex, normal_to_face, low_color, high_color, where, phongmap,
			x1, y1,
			x2, y2,
			x3, y3,
			x4, y4 );
            
        }else{
                MoveFirstPer();
                
                temp = begin_per;  
                while( temp != NULL ){
                        if( count == 0 ){
                                px1 = temp->normal->x;
                                py1 = temp->normal->y;
                        }
                        if( count == 1 ){
                                px2 = temp->normal->x;
                                py2 = temp->normal->y;
                        }
                        if( count == 2 ){
                                px3 = temp->normal->x;
                                py3 = temp->normal->y;
                        }
                        if( count == 3 ){
                                px4 = temp->normal->x;
                                py4 = temp->normal->y;
                        }   
                        ++count;
                        temp = temp->next;
                }

                MoveFirstPer();

		x1 = (int)((px1*128)+127);
		y1 = (int)((py1*128)+127);
		x2 = (int)((px2*128)+127);
		y2 = (int)((py2*128)+127);
		x3 = (int)((px3*128)+127);
		y3 = (int)((py3*128)+127);
		x4 = (int)((px4*128)+127);
		y4 = (int)((py4*128)+127);

		if( x1 < 0 ) x1 = 0;
		if( x2 < 0 ) x2 = 0;
		if( x3 < 0 ) x3 = 0;
		if( x4 < 0 ) x4 = 0;

		if( x1 > 255 ) x1 = 255;
		if( x2 > 255 ) x2 = 255;
		if( x3 > 255 ) x3 = 255;
		if( x4 > 255 ) x4 = 255;

		if( y1 < 0 ) y1 = 0;
		if( y2 < 0 ) y2 = 0;
		if( y3 < 0 ) y3 = 0;
		if( y4 < 0 ) y4 = 0;

		if( y1 > 255 ) y1 = 255;
		if( y2 > 255 ) y2 = 255;
		if( y3 > 255 ) y3 = 255;
		if( y4 > 255 ) y4 = 255;
                          
        	ScanFillFastPhong( in_perspective, normal_to_face, low_color, high_color, where, phongmap,
			x1, y1,
			x2, y2,
			x3, y3,
			x4, y4 );
        }

}

void Face :: DrawSolidTransparentShade( Image *where ){
	int color;

	color = (int)((high_color-low_color)*DotProduct(normal_to_face,light)) + low_color;
	if( color < low_color ) color = low_color;
	if( color > high_color ) color = high_color;

        if( use_perspective == 0.0 ){
		MoveFirst();
	        ScanFillTransparent( vertex, normal_to_face, color, low_color, high_color, where );
	}
	else{
		MoveFirstPer();
        	ScanFillTransparent( in_perspective, normal_to_face, color, low_color, high_color, where );
	}
}

void Face :: DrawSolidFlatShade( Image *where ){
	int color;

	color = (int)((high_color-low_color)*DotProduct(normal_to_face,light)) + low_color;
	if( color < low_color ) color = low_color;
	if( color > high_color ) color = high_color;
	
        if( use_perspective == 0.0 ){
		MoveFirst();
	        ScanFill( vertex, normal_to_face, color, where );
	}
	else{
		MoveFirstPer();
        	ScanFill( in_perspective, normal_to_face, color, where );
	}
}

void Face :: DrawSolidGouraudShade( Image *where ){
        if( use_perspective == 0.0 ){
		MoveFirst();
	        ScanFillGouraud( vertex, normal_to_face, low_color, high_color, where );
	}
	else{
		MoveFirstPer();
        	ScanFillGouraud( in_perspective, normal_to_face, low_color, high_color, where );
	}
}

void Face :: DrawSolidPhongShade( Image *where ){
}


void Face :: RotateX( float angle ){
	matrix_type rotate;
	matrix_type xyzw;
	matrix_type *result;
        float xr, yr, zr, wr;

	xyzw.rows = 4;
	xyzw.cols = 1;

	rotate.rows = rotate.cols = 4;

	rotate.matrix[0][0] = 1.0;
	rotate.matrix[0][1] = rotate.matrix[0][2] = rotate.matrix[0][3] = 0.0;

        rotate.matrix[1][0] = 0.0;
        rotate.matrix[1][1] = cos(angle);
        rotate.matrix[1][2] = -1.0*sin(angle);
	rotate.matrix[1][3] = 0.0;

	rotate.matrix[2][0] = 0.0;
        rotate.matrix[2][1] = sin(angle);
        rotate.matrix[2][2] = cos(angle);
	rotate.matrix[2][3] = 0.0;

	rotate.matrix[3][0] = rotate.matrix[3][1] = rotate.matrix[3][2] = 0.0;
	rotate.matrix[3][3] = 1.0;

        MoveFirst();

        while( vertex != NULL ){
                xyzw.matrix[0][0] = xr = vertex->x;
                xyzw.matrix[1][0] = yr = vertex->y;
                xyzw.matrix[2][0] = zr = vertex->z;
                xyzw.matrix[3][0] = wr = vertex->w;

                result = MatrixMultiply4x4_X_4x1( &rotate, &xyzw );

                vertex->x = (float)result->matrix[0][0];
                vertex->y = (float)result->matrix[1][0];
                vertex->z = (float)result->matrix[2][0];
                vertex->w = (float)result->matrix[3][0];

                xyzw.matrix[0][0] = xr = vertex->normal->x;
                xyzw.matrix[1][0] = yr = vertex->normal->y;
                xyzw.matrix[2][0] = zr = vertex->normal->z;
                xyzw.matrix[3][0] = wr = 1;

                result = MatrixMultiply4x4_X_4x1( &rotate, &xyzw );

                vertex->normal->x = (float)result->matrix[0][0];
                vertex->normal->y = (float)result->matrix[1][0];
                vertex->normal->z = (float)result->matrix[2][0];
                vertex->normal->w = (float)result->matrix[3][0];

                MoveNext();
	}

        xyzw.matrix[0][0] = normal_to_face->x;
        xyzw.matrix[1][0] = normal_to_face->y;
        xyzw.matrix[2][0] = normal_to_face->z;
        xyzw.matrix[3][0] = normal_to_face->w;

        result = MatrixMultiply4x4_X_4x1( &rotate, &xyzw );

        normal_to_face->x = (float)result->matrix[0][0];
        normal_to_face->y = (float)result->matrix[1][0];
        normal_to_face->z = (float)result->matrix[2][0];
        normal_to_face->w = (float)result->matrix[3][0];

	MoveFirst();
}

void Face :: RotateY( float angle){
	matrix_type rotate;
	matrix_type xyzw;
	matrix_type *result;
        float xr, yr, zr, wr;

	xyzw.rows = 4;
	xyzw.cols = 1;

	rotate.rows = rotate.cols = 4;

        rotate.matrix[0][0] = cos(angle);
	rotate.matrix[0][1] = 0.0;
        rotate.matrix[0][2] = sin(angle);
	rotate.matrix[0][3] = 0.0;

	rotate.matrix[1][0] = 0.0;
	rotate.matrix[1][1] = 1.0;
	rotate.matrix[1][2] = rotate.matrix[1][3] = 0.0;

        rotate.matrix[2][0] = -1.0*sin(angle);
	rotate.matrix[2][1] = 0.0;
        rotate.matrix[2][2] = cos(angle);
	rotate.matrix[2][3] = 0.0;

	rotate.matrix[3][0] = rotate.matrix[3][1] = rotate.matrix[3][2] = 0.0;
	rotate.matrix[3][3] = 1.0;

        MoveFirst();        
        while( vertex != NULL ){
                xyzw.matrix[0][0] = xr = vertex->x;
                xyzw.matrix[1][0] = yr = vertex->y;
                xyzw.matrix[2][0] = zr = vertex->z;
                xyzw.matrix[3][0] = wr = vertex->w;
		
		result = MatrixMultiply4x4_X_4x1( &rotate, &xyzw );

                vertex->x = result->matrix[0][0];
                vertex->y = result->matrix[1][0];
                vertex->z = result->matrix[2][0];
                vertex->w = result->matrix[3][0];

                xyzw.matrix[0][0] = xr = vertex->normal->x;
                xyzw.matrix[1][0] = yr = vertex->normal->y;
                xyzw.matrix[2][0] = zr = vertex->normal->z;
                xyzw.matrix[3][0] = wr = 1;

                result = MatrixMultiply4x4_X_4x1( &rotate, &xyzw );


                vertex->normal->x = (float)result->matrix[0][0];
                vertex->normal->y = (float)result->matrix[1][0];
                vertex->normal->z = (float)result->matrix[2][0];
                vertex->normal->w = (float)result->matrix[3][0];

                MoveNext();
	}

        xyzw.matrix[0][0] = normal_to_face->x;
        xyzw.matrix[1][0] = normal_to_face->y;
        xyzw.matrix[2][0] = normal_to_face->z;
        xyzw.matrix[3][0] = normal_to_face->w;

        result = MatrixMultiply4x4_X_4x1( &rotate, &xyzw );

        normal_to_face->x = (float)result->matrix[0][0];
        normal_to_face->y = (float)result->matrix[1][0];
        normal_to_face->z = (float)result->matrix[2][0];
        normal_to_face->w = (float)result->matrix[3][0];

	MoveFirst();
}

void Face :: RotateZ( float angle ){
	matrix_type rotate;
	matrix_type xyzw;
	matrix_type *result;
        float xr, yr, zr, wr;

	xyzw.rows = 4;
	xyzw.cols = 1;

	rotate.rows = rotate.cols = 4;

        rotate.matrix[0][0] = cos(angle);
        rotate.matrix[0][1] = -1.0*sin(angle);
	rotate.matrix[0][2] = rotate.matrix[0][3] = 0.0;

        rotate.matrix[1][0] = sin(angle);
        rotate.matrix[1][1] = cos(angle);
	rotate.matrix[1][2] = rotate.matrix[1][3] = 0.0;

	rotate.matrix[2][0] = rotate.matrix[2][1] = 0.0;
	rotate.matrix[2][2] = 1.0;
	rotate.matrix[2][3] = 0.0;

	rotate.matrix[3][0] = rotate.matrix[3][1] = rotate.matrix[3][2] = 0.0;
	rotate.matrix[3][3] = 1.0;

        MoveFirst();        
        while( vertex != NULL ){
                xyzw.matrix[0][0] = xr = vertex->x;
                xyzw.matrix[1][0] = yr = vertex->y;
                xyzw.matrix[2][0] = zr = vertex->z;
                xyzw.matrix[3][0] = wr = vertex->w;
		
		result = MatrixMultiply4x4_X_4x1( &rotate, &xyzw );

                vertex->x = result->matrix[0][0];
                vertex->y = result->matrix[1][0];
                vertex->z = result->matrix[2][0];
                vertex->w = result->matrix[3][0];

                xyzw.matrix[0][0] = xr = vertex->normal->x;
                xyzw.matrix[1][0] = yr = vertex->normal->y;
                xyzw.matrix[2][0] = zr = vertex->normal->z;
                xyzw.matrix[3][0] = wr = 1;

                result = MatrixMultiply4x4_X_4x1( &rotate, &xyzw );

                vertex->normal->x = (float)result->matrix[0][0];
                vertex->normal->y = (float)result->matrix[1][0];
                vertex->normal->z = (float)result->matrix[2][0];
                vertex->normal->w = (float)result->matrix[3][0];

                MoveNext();
        }

        xyzw.matrix[0][0] = normal_to_face->x;
        xyzw.matrix[1][0] = normal_to_face->y;
        xyzw.matrix[2][0] = normal_to_face->z;
        xyzw.matrix[3][0] = normal_to_face->w;

        result = MatrixMultiply4x4_X_4x1( &rotate, &xyzw );

        normal_to_face->x = (float)result->matrix[0][0];
        normal_to_face->y = (float)result->matrix[1][0];
        normal_to_face->z = (float)result->matrix[2][0];
        normal_to_face->w = (float)result->matrix[3][0];

	MoveFirst();
}
                     
void Face :: AddPerspective( float d ){
        MoveFirst();
        MoveFirstPer();

	in_front_of_camera = 0;

        while( vertex != NULL ){
		if( vertex->z < 0.0 ){
			in_front_of_camera = 1;
			in_perspective->x = ((vertex->x * d) / (vertex->z /*- 1.0*/));
			in_perspective->y = ((vertex->y * d) / (vertex->z /*- 1.0*/));
		}else{
			in_perspective->x = (vertex->x * (32.0));
			in_perspective->y = (vertex->y * (32.0));
		}
		in_perspective->z = vertex->z;
	
		in_perspective->normal = vertex->normal;
		in_perspective->texture_normal = vertex->texture_normal;

                MoveNext();
                MoveNextPer();
	}
}

/*
void  Face :: AddPerspectiveOld( float d ){
	matrix_type perspective;
	matrix_type xyzw;
	matrix_type *result;

	xyzw.rows = 4;
	xyzw.cols = 1;

	perspective.rows = perspective.cols = 4;

	perspective.matrix[0][0] = 1.0;
	perspective.matrix[0][1] = perspective.matrix[0][2] = perspective.matrix[0][3] = 0.0;

        perspective.matrix[1][0] = 0.0;
	perspective.matrix[1][1] = 1.0;
        perspective.matrix[1][2] = 0.0;
        perspective.matrix[1][3] = 0.0;

        perspective.matrix[2][0] = 0.0;
        perspective.matrix[2][1] = 0.0;
	perspective.matrix[2][2] = 1.0;
        perspective.matrix[2][3] = 0.0;

        perspective.matrix[3][0] = perspective.matrix[3][1] = 0.0;
        if( (char)d != 0 )
                perspective.matrix[3][2] = 1.0/d;
        else
                perspective.matrix[3][2] = 1.0;                

	perspective.matrix[3][3] = 1.0;

//        Error("<< %f %f >>",                perspective.matrix[3][2], d );


        MoveFirst();
        MoveFirstPer();
        while( vertex != NULL ){
                xyzw.matrix[0][0] = vertex->x;
                xyzw.matrix[1][0] = vertex->y;
                xyzw.matrix[2][0] = vertex->z;
                xyzw.matrix[3][0] = 1.0;
		
		result = MatrixMultiply4x4_X_4x1( &perspective, &xyzw );

                if( result->matrix[3][0] != 0.0 ){
                        in_perspective->x = result->matrix[0][0]/result->matrix[3][0];
                        in_perspective->y = result->matrix[1][0]/result->matrix[3][0];
                }
                else{
                        in_perspective->x = result->matrix[0][0];
                        in_perspective->y = result->matrix[1][0];
                }
                in_perspective->z = result->matrix[2][0];
                in_perspective->w = result->matrix[3][0];

		in_perspective->normal = vertex->normal;
		in_perspective->texture_normal = vertex->texture_normal;

                MoveNext();
                MoveNextPer();
	}
}
*/

void Face :: Move( float dx, float dy, float dz ){
        MoveFirst();
        MoveFirstPer();

        while( vertex != NULL ){
                vertex->x += dx;
                vertex->y += dy;
                vertex->z += dz;

                in_perspective->x += dx;
                in_perspective->y += dy;
                in_perspective->z += dz;

                MoveNext();
                MoveNextPer();
	}
        
#ifdef PRECALC_NORMALS
        normal_to_face->x += dx;
        normal_to_face->y += dy;
        normal_to_face->z += dz;
#endif       

	MoveFirst(); 
	MoveFirstPer(); 
}

void Normalize( Vertex *vector ){
        float magnitude=(sqrt((vector->x*vector->x)+(vector->y*vector->y)+(vector->z*vector->z)));

        if( magnitude != 0.0 ){
		vector->x = vector->x/magnitude;
		vector->y = vector->y/magnitude;
		vector->z = vector->z/magnitude;
	}
}
                       
Vertex * Face :: FindNormal( void ){
	int i;
        Vertex *normal = NULL;
        
        MoveFirst();
  	/* This is newells methods */
        if( !IsEmpty() ){
                normal = new Vertex;

                normal->x = normal->y = normal->z = 0.0;

                while( vertex->next != NULL ){
                        normal->x += (vertex->y-vertex->next->y)*(vertex->z+vertex->next->z);
                        normal->y += (vertex->z-vertex->next->z)*(vertex->x+vertex->next->x);
                        normal->z += (vertex->x-vertex->next->x)*(vertex->y+vertex->next->y);

                        MoveNext();
		}

                normal->x += (vertex->y-begin->y)*(vertex->z+begin->z);
                normal->y += (vertex->z-begin->z)*(vertex->x+begin->x);
                normal->z += (vertex->x-begin->x)*(vertex->y+begin->y);
        }
         
 	normal_to_face = normal;
        
	return( normal );
}

Vertex * Face :: FindNormalPer( void ){
	int i;
        Vertex *normal = NULL;
        
        MoveFirstPer();
  	/* This is newells methods */
        if( !IsEmpty() ){
                normal = new Vertex;

                normal->x = normal->y = normal->z = 0.0;

                while( in_perspective->next != NULL ){
                        normal->x += (in_perspective->y-in_perspective->next->y)*(in_perspective->z+in_perspective->next->z);
                        normal->y += (in_perspective->z-in_perspective->next->z)*(in_perspective->x+in_perspective->next->x);
                        normal->z += (in_perspective->x-in_perspective->next->x)*(in_perspective->y+in_perspective->next->y);

                        MoveNextPer();
		}

                normal->x += (in_perspective->y-begin_per->y)*(in_perspective->z+begin_per->z);
                normal->y += (in_perspective->z-begin_per->z)*(in_perspective->x+begin_per->x);
                normal->z += (in_perspective->x-begin_per->x)*(in_perspective->y+begin_per->y);
        }

	return( normal );
}


char Face :: Visible( void ){
        Vertex *normal;
        char visible = 0;

        normal = normal_to_face;

        /* change this to < for z increasing into screen > for decreasing */
        /* < for polygon order counter-clockwise > for clockwise */
        if( normal->z <= 0.0 )   
                visible = 1;
        
	return( visible );
}

void Face :: UseTexture( char *filename ){
        use_texture=1;

	LoadPCX( filename, (unsigned char **)&(texture.bitmap), &palette, &(texture.height), &(texture.width) );
}

void Face :: UseStretchedTexture( char *filename ){
        use_stretched_texture=1;

	LoadPCX( filename, (unsigned char **)&(texture.bitmap), &palette, &(texture.height), &(texture.width) );
}

           
/*** Object_3d class functions ***/

void Object_3d :: CalculateNormals( void ){
        Vertex *temp;
        MoveFirst();

        while( face != NULL ){
                temp = new Vertex;
                temp = face->face->FindNormal();
                Normalize( temp );
                face->face->normal_to_face = temp;
                MoveNext();
        }

	MoveFirst();
}

char VertexExists( int id, Vertex *vertex ){
        while( vertex != NULL ){
                if( vertex->id == id ){
                        return( 1 );
                }
                vertex = vertex->next;
        }

        return( 0 );
}


/** goes through and assigns like x,y,z 's the same id **/

void Object_3d :: GroupVertexIDs( void ){
	int id;
	float x, y, z;
	Face_list *fbegin;
	Vertex *temp_search;
	Face_list *temp_face;
	Vertex *temp_vertex;

	MoveFirst();
	fbegin = face;

	/* outer two loops go through each vertex, inner 2 to go through each and compare */

	while( face != NULL ){
		temp_search = face->face->begin; /* first vertex */
		while( temp_search != NULL ){
			x = temp_search->x;
			y = temp_search->y;
			z = temp_search->z;
			id = temp_search->id;
			temp_face = fbegin;
			while( temp_face != NULL ){
				temp_vertex = temp_face->face->begin;  /* first vertex */
				while( temp_vertex != NULL ){
					if( temp_vertex->x == x && temp_vertex->y == y && temp_vertex->z == z ){
						temp_vertex->id = id;
					}
					temp_vertex = temp_vertex->next;
				}
			temp_face = temp_face->next;
			}
				
			temp_search = temp_search->next;
                }
                MoveNext();
        }

	MoveFirst();
}


/* just a debugging function */

void Object_3d :: ShowGroupVertexIDs( void ){
	int id;
	float x, y, z;
	Face_list *fbegin;
	Vertex *temp_search;
	Face_list *temp_face;
	Vertex *temp_vertex;

	MoveFirst();
	fbegin = face;

	/* outer two loops go through each vertex, inner 2 to go through each and compare */

	while( face != NULL ){
		temp_search = face->face->begin; /* first vertex */
		while( temp_search != NULL ){
			printf( "(%lf, %lf, %lf) -> %d\n", temp_search->x, temp_search->y, temp_search->z, temp_search->id);
			temp_search = temp_search->next;
                }
                MoveNext();
        }

	MoveFirst();
}


/* calculate the normals (and store it) for each vertex in object */  

void Object_3d :: CalculateNormalsForVertices( void ){
        int id;
        float normx=0.0, normy=0.0, normz=0.0;
        int count=0;
        Vertex *temp_search;
        Vertex *normal;
        Face_list *temp_face, *fbegin;

	CalculateNormals();  /* doing this here will save 3 calculations per vertex */

        MoveFirst();
        fbegin = face;

	while( face != NULL ){
		temp_search = face->face->begin; /* first vertex */
		while( temp_search != NULL ){
			id = temp_search->id;
			normx = normy = normz = 0.0;
			count = 0;	/* number of adjacent faces */
			temp_face = fbegin;
			while( temp_face != NULL ){
				/* if the vertex exists in the face then add it to be averaged */
				if( VertexExists( id, temp_face->face->begin) ){
                              		normx += (float)temp_face->face->normal_to_face->x;
                                       	normy += (float)temp_face->face->normal_to_face->y;
                                       	normz += (float)temp_face->face->normal_to_face->z;
					++count;
				}
				temp_face = temp_face->next;
			}

                        if( count != 0 ){	/* finish up average */
                                normx = normx / (float)count;
                                normy = normy / (float)count;
                                normz = normz / (float)count;
                        }
                        normal = NewVertex( normx, normy, normz );
                        Normalize( normal );
                        temp_search->normal = normal;	/* store the normal to vertex */
                        temp_search->texture_normal = normal;	/* store the normal to vertex */
				
			temp_search = temp_search->next;
                }
                MoveNext();
        }

	MoveFirst();
}

char ContainsFloatSlash( char *str ){
        int i=0;

        if( str != NULL ){
          while( str[i] != '\0' ){
                if( str[i] == '/' )
                        if( str[i+1] == '/' )
                               return( 1 );
                ++i;
          }
        }

        return( 0 );
}

char *ReturnFirstPart( char *str ){
        char *temp=new char[99];
        int i=0;

        while( str[i] != '/' ){
                temp[i] = str[i];
                ++i;
        }

        temp[i] = '\0';

        return( temp );
}

char *ReturnLastPart( char *str ){
        char *temp=new char[99];
        int i=0;
        int j=0;

        while( str[i] != '/' ){
                ++i;
        }
        i += 2;

        while( str[i] != '\0' ){
                temp[j] = str[i];
                ++i;
                ++j;
        }

        temp[j] = '\0';

        return( temp );
}

void Object_3d :: UseNoShade( void ){
        use_no_shade=1;
        use_flat_shade=0;
        use_gouraud_shade=0;
        use_phong_shade=0;
        use_fastphong_shade=0;
        use_transparent_shade=0;
        use_texture=0;
}

void Object_3d :: UseFlatShade( void ){
        use_no_shade=0;
        use_flat_shade=1;
        use_gouraud_shade=0;
        use_phong_shade=0;
        use_fastphong_shade=0;
        use_transparent_shade=0;
        use_texture=0;
}

void Object_3d :: UseGouraudShade( void ){
        use_no_shade=0;
        use_flat_shade=0;
        use_gouraud_shade=1;
        use_phong_shade=0;
        use_fastphong_shade=0;
        use_transparent_shade=0;
        use_texture=0;
}

void Object_3d :: UsePhongShade( void ){
        use_no_shade=0;
        use_flat_shade=0;
        use_gouraud_shade=0;
        use_phong_shade=1;
        use_fastphong_shade=0;
        use_transparent_shade=0;
        use_texture=0;
}

void Object_3d :: UseFastPhongShade( void ){
        use_no_shade=0;
        use_flat_shade=0;
        use_gouraud_shade=0;
        use_phong_shade=0;
        use_fastphong_shade=1;
        use_transparent_shade=0;
        use_texture=0;

	CalcPhongMap();
}

void Object_3d :: UseTransparentShade( void ){
        use_no_shade=0;
        use_flat_shade=0;
        use_gouraud_shade=0;
        use_phong_shade=0;
        use_fastphong_shade=0;
        use_transparent_shade=1;
        use_texture=0;
}

void Object_3d :: UseTexture( char *filename ){
        use_no_shade=0;
        use_flat_shade=0;
        use_gouraud_shade=0;
        use_phong_shade=0;
        use_fastphong_shade=0;
        use_transparent_shade=0;
        use_texture=1;
	unsigned char *tempbitmap;
	int tempwidth;
	int tempheight;
	paletteType temppal;

	MoveFirst();

	LoadPCX( filename, (unsigned char **)&tempbitmap, &temppal, &tempheight, &tempwidth );

	while( face != NULL ){
		face->face->use_texture = 1;
		face->face->texture.bitmap = tempbitmap;
		face->face->texture.width  = tempwidth;
		face->face->texture.height = tempheight;
		memcpy( &(face->face->palette), &temppal, sizeof( paletteType ) );

		MoveNext();
        }

	MoveFirst();
}

void Object_3d :: UseTexture( Image *image ){
        use_no_shade=0;
        use_flat_shade=0;
        use_gouraud_shade=0;
        use_phong_shade=0;
        use_fastphong_shade=0;
        use_transparent_shade=0;
        use_texture=1;

	MoveFirst();

	while( face != NULL ){
		face->face->use_texture = 1;
		face->face->texture.bitmap = image->bitmap;
		face->face->texture.width  = image->width;
		face->face->texture.height = image->height;

		MoveNext();
        }

	MoveFirst();
}

void Object_3d :: UseStretchedTexture( char *filename ){
        use_no_shade=0;
        use_flat_shade=0;
        use_gouraud_shade=0;
        use_phong_shade=0;
        use_fastphong_shade=0;
        use_transparent_shade=0;
        use_texture=0;
        use_stretched_texture=1;
	unsigned char *tempbitmap;
	int tempwidth;
	int tempheight;
	paletteType temppal;

	MoveFirst();

	LoadPCX( filename, (unsigned char **)&tempbitmap, &temppal, &tempheight, &tempwidth );

	while( face != NULL ){
		face->face->use_stretched_texture = 1;
		face->face->texture.bitmap = tempbitmap;
		face->face->texture.width  = tempwidth;
		face->face->texture.height = tempheight;
		memcpy( &(face->face->palette), &temppal, sizeof( paletteType ) );

		MoveNext();
        }

	MoveFirst();
}

void Object_3d :: UseStretchedTexture( Image *image ){
        use_no_shade=0;
        use_flat_shade=0;
        use_gouraud_shade=0;
        use_phong_shade=0;
        use_fastphong_shade=0;
        use_transparent_shade=0;
        use_texture=0;
        use_stretched_texture=1;


	MoveFirst();

	while( face != NULL ){
		face->face->use_stretched_texture = 1;
		face->face->texture.bitmap = image->bitmap;
		face->face->texture.width  = image->width;
		face->face->texture.height = image->height;

		MoveNext();
        }

	MoveFirst();
}

 
Object_3d :: Object_3d( char *filename ){
        FILE *fp;
        char **parsedstr;
        char tempstr[199];

        Face *vertices;     /* make one long list of vertices */
        Face *normals;      /* make one long list of normals  */

        Face *temp;     /* used for adding new faces */

	char reset_vertex_id = 0;

        vertices = new Face();
        normals  = new Face();

	cur_vertex_id = 1;
        cur_normal_id = 1;

        begin = face = end = NULL;      /* initialize stuff */

	use_perspective = 0.0;
	remove_backfaces = 0;

	low_color = 0;
	high_color = 255;

        use_no_shade=0;
        use_flat_shade=1;
        use_gouraud_shade=0;
        use_phong_shade=0;
        use_fastphong_shade=0;
        use_transparent_shade=0;

	phongmap = NULL;
        
        if( (fp = fopen( filename, "r" )) == NULL )
                Error( "Couldn't open '%s'.", filename );

        while( !feof( fp ) ){
                tempstr[0]='\0';
                fgets( tempstr, 199, fp );
                parsedstr = Parse( tempstr );

                int i=0;
                if( parsedstr != NULL ){
                        if( strcmp( parsedstr[i], "v" ) == 0 ){
                                vertices->AddVertex(    ((atof( parsedstr[1] )-VIEWPORT_MIN_X)/(VIEWPORT_MAX_X-VIEWPORT_MIN_X)),
                                                        ((atof( parsedstr[2] )-VIEWPORT_MIN_Y)/(VIEWPORT_MAX_Y-VIEWPORT_MIN_Y)),
                                                        ((atof( parsedstr[3] )-VIEWPORT_MIN_Z)/(VIEWPORT_MAX_Y-VIEWPORT_MIN_Z)) );
                                                        
                        }
                        else if( strcmp( parsedstr[i], "vn" ) == 0 ){
                                normals->AddVertex(     ((atof( parsedstr[1] )-VIEWPORT_MIN_X)/(VIEWPORT_MAX_X-VIEWPORT_MIN_X)),
                                                        ((atof( parsedstr[2] )-VIEWPORT_MIN_Y)/(VIEWPORT_MAX_Y-VIEWPORT_MIN_Y)),
                                                        ((atof( parsedstr[3] )-VIEWPORT_MIN_Z)/(VIEWPORT_MAX_Y-VIEWPORT_MIN_Z)) );
                        }
                        else if( strcmp( parsedstr[i], "f" ) == 0 ){
				
				/* do this cause the AddVertex()s above inc'd the counter */	
				if( !reset_vertex_id ){
					reset_vertex_id = 1;
					cur_vertex_id = 1;
				}
                                temp = new Face();

                                ++i;
                                while( parsedstr[i] != NULL ){
                                        if( ContainsFloatSlash( parsedstr[i] ) ){
                                                if( vertices->MoveToVertex( atoi( ReturnFirstPart(parsedstr[i]) ) ) == 0)
                                                        Error( "Cannot find vertex %s >> %s %s %s %s\n", parsedstr[i], ReturnFirstPart(parsedstr[0]), ReturnFirstPart(parsedstr[1]), ReturnFirstPart(parsedstr[2]), ReturnFirstPart(parsedstr[3]) );
                                                //if( normals->MoveToVertex( atoi( ReturnLastPart(parsedstr[i]) ) ) == 0)
                                                //        Error( "Cannot find normal %s >> %s %s %s %s\n", parsedstr[i], parsedstr[0], ReturnLastPart(parsedstr[1]), ReturnLastPart(parsedstr[2]), ReturnLastPart(parsedstr[3]) );
                                                
                                                temp->AddVertex( vertices->vertex->x, vertices->vertex->y, vertices->vertex->z );
                                                //temp->AddVertex( vertices->vertex->x, vertices->vertex->y, vertices->vertex->z, normals->vertex->x, normals->vertex->y, normals->vertex->z );
                                                ++i;        
                                        }else{
						if( vertices->MoveToVertex( atoi( parsedstr[i] ) ) == 0)
                                                        Error( "Cannot find vertex %s >> %s %s %s %s\n", parsedstr[i], parsedstr[0], parsedstr[1], parsedstr[2], parsedstr[3] );
                                                temp->AddVertex( vertices->vertex->x, vertices->vertex->y, vertices->vertex->z );
                                                ++i;
                                        }
                                }

                                temp->MoveFirst();

                                AddFace( temp );
                        }
                        else{
                                printf("Command ignored: %s\n", parsedstr[i] );
                        }
                }
        }

        fclose( fp );

	GroupVertexIDs();
        CalculateNormalsForVertices();

/*** uncomment me to use slow phong shading 
        printf("Calculating normals to vertices...");fflush(stdout);
        if( face->face->vertex->normal == NULL )
                CalculateNormalsForVertices();
        printf("done.\n");fflush(stdout);
        MoveFirst();
***/
        float highestx=-999999.0, lowestx=999999.0;
        float highesty=-999999.0, lowesty=999999.0;
        float highestz=-999999.0, lowestz=999999.0;
        while( face != NULL ){
                face->face->MoveFirst();
                while( face->face->vertex != NULL ){
			if( face->face->vertex->x > highestx )
				highestx = face->face->vertex->x;
			if( face->face->vertex->x < lowestx )
				lowestx = face->face->vertex->x;

			if( face->face->vertex->y > highesty )
				highesty = face->face->vertex->y;
			if( face->face->vertex->y < lowesty )
				lowesty = face->face->vertex->y;

			if( face->face->vertex->z > highestz )
				highestz = face->face->vertex->z;
			if( face->face->vertex->z < lowestz )
				lowestz = face->face->vertex->z;

                        //printf("%d  %f, %f, %f\n", face->face->vertex->id,face->face->vertex->normal->x,face->face->vertex->normal->y, face->face->vertex->normal->z);
                        face->face->MoveNext();
                }
                MoveNext();
        }
	printf("Highest: x=%lf  y=%lf  z=%lf\n", highestx, highesty, highestz );
	printf("Lowest: x=%lf  y=%lf  z=%lf\n", lowestx, lowesty, lowestz );
        MoveFirst();
}


Object_3d :: Object_3d( void ){
        begin = face = end = NULL;

        use_perspective = 0.0;
        remove_backfaces = 0;
}

Object_3d :: ~Object_3d( void ){
        /* free memory */
}

void Object_3d :: UsePerspective( float v ){
        MoveFirst();
        while( face != NULL ){
                face->face->UsePerspective( -2.0 );
                MoveNext();
        }
        use_perspective = -2.0;
}

void Object_3d :: RemoveBackfaces( char v ){
        MoveFirst();
        while( face != NULL ){
                face->face->RemoveBackfaces( v );
                MoveNext();
        }
        remove_backfaces = v;
}

void Object_3d :: MoveFirst( void ){
        face = begin;
}

char Object_3d :: MoveNext( void ){
        char result = 0;

        face = face->next;

        if( face != NULL )
                result = 1;

        return( result );
}

void Object_3d :: MoveLast( void ){
        face = end;
}

void Object_3d :: AddFace( Face *add ){
        Face_list *temp;

        //temp = new Face_list;
        temp = new Face_list;

        temp->face = add;
        temp->next = NULL;

        if( IsEmpty() ){
                begin = face = end = temp;
        }else{
                end->next = temp;
                end = temp;
        }
}

void Object_3d :: Remove( void ){
        /* do removal stuff and stuff */
}

char Object_3d :: IsEmpty( void ){
        return( begin == NULL );
}
                      
char *GetWord( char *str, int *pos ){
        static char tempstr[99];
        int j=0;

        for(; str[*pos] == ' '; ++(*pos) );
        for(j=0; str[*pos] != ' ' && str[*pos] != '\0'; ++(*pos), ++j ){
                if( str[*pos] != '\n' )
                        tempstr[j] = str[*pos];
                        tempstr[j+1] = '\0';            /* temp fix */
        }

        tempstr[j] = '\0';

        return( tempstr );
}

char **Parse( char *str ){
        char **tempstr;
        char *tempword;
        int i = 0;
        int word_count = 0;

        for(int cv=0; str[cv] != '\0'; ++cv ){
                if( str[cv] == '\n' )
                        str[cv] = '\0';
        }

        /* count words */
        while( str[i] != '\0' ){
                tempword = GetWord( str, &i );
                if( tempword[0] != '\0' ){
                        ++word_count;
                }
        }

        if( word_count == 0 ){
                tempstr = NULL;
        }else{
                tempstr = (char **)malloc( sizeof( char * )*(word_count+1));
                for(int index=0; index<(word_count+1); ++index )
                        tempstr[index] = (char *)malloc( sizeof( char ) * 99 );

                i = 0;

                for(int index=0; index<word_count; ++index ){
                        strcpy( tempstr[index], GetWord( str, &i ) );
                }
             
                // printf("<<%s>>\n", tempstr[word_count-1] );
             
                tempstr[word_count] = NULL;
                
        }

        return( tempstr );
}

void Object_3d :: SetCurrentAsOrig( void ){
	MoveFirst();


	while( face != NULL ){
		face->face->MoveFirst();

		if( face->face->normal_to_face != NULL ){
			face->face->normal_to_face->orig_x = face->face->normal_to_face->x;
			face->face->normal_to_face->orig_y = face->face->normal_to_face->y;
			face->face->normal_to_face->orig_z = face->face->normal_to_face->z;
			face->face->normal_to_face->orig_w = face->face->normal_to_face->w;
		}

		while( face->face->vertex != NULL ){ 
			face->face->vertex->orig_x = face->face->vertex->x;
			face->face->vertex->orig_y = face->face->vertex->y;
			face->face->vertex->orig_z = face->face->vertex->z;
			face->face->vertex->orig_w = face->face->vertex->w;

			if( face->face->vertex->normal != NULL ){
				face->face->vertex->normal->orig_x = face->face->vertex->normal->x;
				face->face->vertex->normal->orig_y = face->face->vertex->normal->y;
				face->face->vertex->normal->orig_z = face->face->vertex->normal->z;
				face->face->vertex->normal->orig_w = face->face->vertex->normal->w;
			}
		
			face->face->MoveNext();
		}

		face->face->MoveFirst();	

		MoveNext();
	}

	MoveFirst();
}

void Object_3d :: Reset( void ){
	MoveFirst();


	while( face != NULL ){
		face->face->MoveFirst();
		
		if( face->face->normal_to_face != NULL ){
                        face->face->normal_to_face->x = face->face->normal_to_face->orig_x;
                        face->face->normal_to_face->y = face->face->normal_to_face->orig_y;
                        face->face->normal_to_face->z = face->face->normal_to_face->orig_z;
                        face->face->normal_to_face->w = face->face->normal_to_face->orig_w;
                }

		while( face->face->vertex != NULL ){ 
			face->face->vertex->x = face->face->vertex->orig_x;
			face->face->vertex->y = face->face->vertex->orig_y;
			face->face->vertex->z = face->face->vertex->orig_z;
			face->face->vertex->w = face->face->vertex->orig_w;

			if( face->face->vertex->normal != NULL ){
				face->face->vertex->normal->x = face->face->vertex->normal->orig_x;
				face->face->vertex->normal->y = face->face->vertex->normal->orig_y;
				face->face->vertex->normal->z = face->face->vertex->normal->orig_z;
				face->face->vertex->normal->w = face->face->vertex->normal->orig_w;
			}
		
			face->face->MoveNext();
		}

		face->face->MoveFirst();	

		MoveNext();
	}

	MoveFirst();
}

void Object_3d :: DrawSolid( Image *where ){
	/* instead of calculating the normals each time, we rotate them */
	//CalculateNormalsForVertices();

        if( use_perspective != 0.0 ){
                Move( -0.5, -0.5, -0.5 );
                AddPerspective( use_perspective );
                Move( 0.5, 0.5, 0.5 );
        }

        if( remove_backfaces ){
		Vertex *norm;
		MoveFirst();

                while( face != NULL ){
		        if( use_perspective != 0.0 ){
				norm = face->face->FindNormalPer();
			}else{
				norm = face->face->FindNormal();
			}
			if( (use_perspective != 0.0) && face->face->in_front_of_camera ){
				if( norm->z < 0.00001 ){		/* approximate 0 because of rounding errors */
                                	if( use_no_shade )
                                        	face->face->DrawSolidNoShade( where );
                                	if( use_flat_shade )
                                        	face->face->DrawSolidFlatShade( where );
                                	if( use_gouraud_shade )
                                        	face->face->DrawSolidGouraudShade( where );
                                	if( use_phong_shade )
                                        	face->face->DrawSolidPhongShade( where );
                                	if( use_fastphong_shade )
                                        	face->face->DrawSolidFastPhongShade( where, phongmap );
                                	if( use_transparent_shade )
                                        	face->face->DrawSolidTransparentShade( where );
                                	if( use_texture )
                                        	face->face->DrawSolidTexture( where );
                                	if( use_stretched_texture )
                                        	face->face->DrawSolidStretchedTexture( where );
				}
                        }
                        MoveNext();
                }
        }else{
                MoveFirst();
                while( face != NULL ){
			if( (use_perspective != 0.0) && face->face->in_front_of_camera ){
	                        if( use_no_shade )
        	                        face->face->DrawSolidNoShade( where );
                	        if( use_flat_shade )
                	             face->face->DrawSolidFlatShade( where );
				if( use_gouraud_shade )
                                	face->face->DrawSolidGouraudShade( where );
                        	if( use_phong_shade )
                                	face->face->DrawSolidPhongShade( where );
                        	if( use_fastphong_shade )
                                	face->face->DrawSolidFastPhongShade( where, phongmap );
                        	if( use_transparent_shade )
                                	face->face->DrawSolidTransparentShade( where );
                        	if( use_texture )
                                	face->face->DrawSolidTexture( where );
                        	if( use_stretched_texture )
                                	face->face->DrawSolidStretchedTexture( where );
			}
                        MoveNext();
                }
        }
}
                                     
void Object_3d :: DrawWireFrame( Image *where ){
        if( (char)use_perspective != 0 ){
                Move( -0.5, -0.5, -0.5 );
                AddPerspective( use_perspective );
                Move( 0.5, 0.5, 0.5 );
        }


        if( remove_backfaces ){
                MoveFirst();

                while( face != NULL ){
                        if( face->face->Visible() ){
                                if( (char)use_perspective != 0 )
                                        face->face->DrawWireFrameInPerspective( where );
                                else
                                        face->face->DrawWireFrameNoPerspective( where );
                        }
                        MoveNext();
                }
        }else{
                if( (char)use_perspective != 0 )
                        DrawWireFrameInPerspective( where );
                else
                        DrawWireFrameNoPerspective( where );
        }
}

void Object_3d :: DrawWireFrameNoPerspective( Image *where ){
        MoveFirst();

        while( face != NULL ){
                face->face->DrawWireFrameNoPerspective( where );
                MoveNext();
        }
}                             

void Object_3d :: DrawWireFrameInPerspective( Image *where ){
        MoveFirst();

        while( face != NULL ){
                face->face->DrawWireFrameInPerspective( where );
                MoveNext();
        }
}                             

void Object_3d :: RotateX( float angle ){
        MoveFirst();

        while( face != NULL ){
                face->face->RotateX( angle );
                MoveNext();
        }
}

void Object_3d :: RotateY( float angle ){
        MoveFirst();

        while( face != NULL ){
                face->face->RotateY( angle );
                MoveNext();
        }
}

void Object_3d :: RotateZ( float angle ){
        MoveFirst();

        while( face != NULL ){
                face->face->RotateZ( angle );
                MoveNext();
        }
}

void Object_3d :: AddPerspective( float d ){
        MoveFirst();

        while( face != NULL ){
                face->face->AddPerspective( d );
                MoveNext();
        }
}                                            

void Object_3d :: Move( float dx, float dy, float dz ){
        MoveFirst();

        while( face != NULL ){
                face->face->Move( dx, dy, dz );
                MoveNext();
        }

	MoveFirst();
}
        
matrix_type *MatrixMultiply4x4_X_4x1( matrix_type *a, matrix_type *b ){
	int i,j;
        float temp = 0.0;
	static matrix_type *result=NULL;

	if( result == NULL ){
        	if( (result = new matrix_type) == NULL )
                	Error( "Ran out of memory!.\n\n" );
	}

	result->rows = a->rows;
	result->cols = b->cols;

	for(j=0; j<4; ++j ){
		for(i=0; i<4; ++i ){
                        temp += a->matrix[j][i]*b->matrix[i][0];
		}
                result->matrix[j][0] = temp;
		temp = 0.0;
	}

	return( result );
}

matrix_type *MatrixMultiply4x4_X_4x4( matrix_type *a, matrix_type *b ){
	int i,j,k;
        float temp = 0.0;
	static matrix_type *result=NULL;

	if( result == NULL ){
	        if( (result = new matrix_type) == NULL )
        	        Error( "Ran out of memory!.\n\n" );
	}

	result->rows = a->rows;
	result->cols = b->cols;

	for(k=0; k<4; ++k){
		for(j=0; j<4; ++j ){
			for(i=0; i<4; ++i ){
                                temp += a->matrix[k][i]*b->matrix[i][j];
			}
                        result->matrix[k][j] = temp;
                        temp = 0.0;
		}
	}
	return( result );
}

