#include <string.h>
#include "graphics.h"
#include "error.h"

#define TRANSPARENCY 0.7

#define Blend( x, y, color, image, percent )   ( (image)->bitmap[ ((y)*(image)->width)+(x) ] * (percent) ) + ( (color) * ( 1.0 - (percent) ) )


extern "C" {
	#include "libsx.h"
	#include "callbacks.h"
}

typedef struct {
	char    manufacturer;
	char    version;
	char    encoding;
	char    bits_per_pixel;
        short   xmin,ymin;
        short   xmax,ymax;
        short   hres;
        short   vres;
	char    pallete[48];
	char    reserved;
	char    colour_planes;
        short   bytes_per_line;
        short   pallete_type;
	char    filler[58];
		} PCXHEAD;
PCXHEAD   header;

/****** use c instead of c++ cause couldn't compile callbacks.c with g++ ****/

void Clear( Image *image ){
        memset( image->bitmap, 0, image->width*image->height );
}

/*
void SetPixel( int x, int y, int color, Image *image  ){
      if( x >= 0 && y >= 0 && x < image->width && y < image->height )
              image->bitmap[ (y*image->width)+x ] = color;
}
*/


void MyDrawLine( int x1, int y1, int x2, int y2, int color, Image *image ){
        int dx, dy;
        int x, y;
        int endx, endy;
        int startx, starty;
        char incrementX = 1;
        double ratio=0.0;

        dx = ( x1 - x2 );
        dy = ( y1 - y2 );
        
        if( abs(dy) > abs(dx) ) incrementX = 0;

        if( incrementX ){
                if( dx != 0 )
                        ratio = (double)dy/(double)dx;
                if( x1 < x2 ){
                        x = x1;
                        y = y1;
                        endx = x2;
			endy = y2;
                        starty = y1;
                        startx = x1;
                }
                else{
                        x = x2;
                        y = y2;
                        endx = x1;
			endy = y1; 
                        starty = y2;
                        startx = x2;
                }
                while( x < endx ){
                        SetPixel(x, y, Blend( x, y, color, image, TRANSPARENCY ), image );
                        ++x;
                        y=(int)(starty+(ratio*(x-startx)));
                }
/*
                SetPixel( startx, starty, Blend( startx, starty, color, image, TRANSPARENCY ), image );
                SetPixel( endx, endy, Blend( endx, endy, color, image, TRANSPARENCY ), image );
*/
        }
        else{
                if( dy != 0 )
                        ratio = (double)dx/(double)dy;

                if( y1 < y2 ){
                        y = y1;
                        x = x1;
                        endy = y2;
			endx = x2;
                        startx = x1;
                        starty = y1;
                }
                else{
                        y = y2;
                        x = x2;
                        endy = y1;
			endx = x1;
                        startx = x2;
                        starty = y2;
                }
                while( y <= endy ){
                        SetPixel(x, y, Blend( x, y, color, image, TRANSPARENCY ), image );
                        ++y;
                        x=(int)(startx+(ratio*(y-starty)));
                }
/*
                SetPixel( startx, starty, Blend( startx, starty, color, image, TRANSPARENCY ), image );
                SetPixel( endx, endy, Blend( endx, endy, color, image, TRANSPARENCY ), image );
*/
        }
}

/***************

void Image :: Clear( void ){
        memset( bitmap, 255, width*height );
}

void Image :: SetPixel( int x, int y, int color ){
      if( x >= 0 && y >= 0 && x < width && y < height )
              bitmap[ (y*width)+x ] = color;
}

void Image :: DrawLine( int x1, int y1, int x2, int y2, int color ){
        int dx, dy;
        int x, y;
        int endx, endy;
        int startx, starty;
        char incrementX = 1;
        double ratio=0.0;

	color = 0;

        dx = ( x1 - x2 );
        dy = ( y1 - y2 );
        
        if( abs(dy) > abs(dx) ) incrementX = 0;

        if( incrementX ){
                if( dx != 0 )
                        ratio = (double)dy/(double)dx;
                if( x1 < x2 ){
                        x = x1;
                        y = y1;
                        endx = x2;
			endy = y2;
                        starty = y1;
                        startx = x1;
                }
                else{
                        x = x2;
                        y = y2;
                        endx = x1;
			endy = y1; 
                        starty = y2;
                        startx = x2;
                }
                while( x < endx ){
                        SetPixel(x, y, color );
                        ++x;
                        y=(int)(starty+(ratio*(x-startx)));
                }
                SetPixel( startx, starty, color );
                SetPixel( endx, endy, color );
        }
        else{
                if( dy != 0 )
                        ratio = (double)dx/(double)dy;

                if( y1 < y2 ){
                        y = y1;
                        x = x1;
                        endy = y2;
			endx = x2;
                        startx = x1;
                        starty = y1;
                }
                else{
                        y = y2;
                        x = x2;
                        endy = y1;
			endx = x1;
                        startx = x2;
                        starty = y2;
                }
                while( y <= endy ){
                        SetPixel(x, y, color );
                        ++y;
                        x=(int)(startx+(ratio*(y-starty)));
                }
                SetPixel(startx, starty, color );
                SetPixel(endx, endy, color );
        }
}


Image :: Image( char *filename ){
        int size,x,c=0,count;
        PCXHEAD header;
	long int n;
        int bytes, i=0;
	long int   inc;
	FILE *fp;

        transparent_color = 0;

        if( (fp = fopen(filename,"rb")) != NULL ){
                  // read in the header
                  if( fread( (char *)&header, 1, sizeof(PCXHEAD), fp) == sizeof(PCXHEAD) ){
		  // check to make sure it's a picture 
                         if( header.manufacturer == 0x0a && header.version == 5 ){
                                 if( !fseek(fp, -769L, SEEK_END) ){
                                         if( fgetc(fp) == 0x0c && fread( palette, 1, 768, fp) == 768){
                                                 fseek( fp, 128L, SEEK_SET) ;
						 //allocate a big buffer 
                                                 width  = (header.xmax - header.xmin) + 1;
                                                 height = (header.ymax - header.ymin) + 1;
                                                 if( (bitmap = (char *)malloc( sizeof(char)*(width*height) )) == NULL )
							 Error( "Ran out of memory.\n" );
						 bytes=header.bytes_per_line;
                                        }else
						 Error("error reading pallete");
				}else
					Error("error seeking to pallete");
			}else
				Error("not a 256 color pcx file.\n");
                }else
                        Error("error reading %s.\n",filename);
        }else
                Error ("Error opening %s.\n",filename);

///////////////////////////////////////////////////////////////////////

        for (i = 0; i < height*width; ++i){
		c=fgetc(fp) & 0xff;  // if it's a run of bytes field 

		if((c & 0xc0) == 0xc0) { // and off the high bits 
                        x = c & 0x3f;
			// run the byte 
                        c = fgetc(fp);
                        while( x-- ){
                                bitmap[i++] = c;
			}
			i--;
                }else{
                        // just store it 
                        bitmap[i] = c;
                }
         }

         // convert eight bits to six bits 
         for( i=0; i<768; ++i )
                palette[i] = palette[i] >> 2;

	 fclose(fp);
}

Image :: Image( int width, int height ){
        transparent_color = 0;

        if( (bitmap = (char *)malloc( sizeof(char)*width*height )) == NULL )
                Error( "Ran out of memory.\n" );

	Clear();
}

Image :: ~Image( void ){
        free( bitmap );
}

void Image :: Draw( int ulx, int uly, Image *where ){
        long x, y, byte=0;
        int ystop=uly+height,
            xstop=ulx+width;

        for( y = uly; y < ystop; ++y ){
                for( x = ulx; x < xstop; ++x ){
                  if( x > -1 && y > -1 && x < 320 && y < 200)
                        if ( bitmap[byte] != transparent_color )
                                where[( x + (y << 8) + (y << 6))] = bitmap[byte];
			++byte;
		}
        }
}


**************/
